
#include <pthread.h>

#include "Clean.h"

#ifdef CLEAN_ORB_MICO
#define MICO_CONF_DYNANY
#define MICO_CONF_POA
#include "CORBA-SMALL.h"
#endif

#ifdef CLEAN_ORB_ORBACUS
#include <OB/CORBA.h>
#endif

#include <map>

#include "show_corba.h"

// We want C linking for all functions callable from Clean
extern "C" {

#define ARRIVED cerr << "Arrived in " << __FILE__ << " at " << __LINE__ << endl; flush(cerr);

#ifdef CLEAN_ORB_MICO
#if MICO_BIN_VERSION < 0x020305
#define NON_CONST_ANY_EXTRACTION
#endif
#endif

////////////////////////////////
//        GLOBALS
////////////////////////////////

// Set by 'corba_orb_init'
static CORBA::ORB_ptr the_orb;

static PortableServer::POA_ptr the_poa;

static DynamicAny::DynAnyFactory_var dynany_factory;

// Useful for debugging
static CORBA_OUT out(cout);

////////////////////////////////
//       UTILITY FUNCTIONS
////////////////////////////////

static char * 
clean_strdup(CleanString value)
{
  char * result;

  result = new char[(CleanStringLength(value) + 1)];
  assert(result);
  memcpy(result, CleanStringCharacters(value), CleanStringLength(value));
  result[CleanStringLength(value)] = '\0';

  return(result);
}

CleanString 
clean_string_create(const char * str)
{
  char * result;
  int len;

  len = strlen(str);
  result = (char*)malloc(len + 16);
  assert(result);

  CleanStringLength(result) = len;
  memcpy(CleanStringCharacters(result), str, len);

  return ((CleanString)result);
}

void set_type_of_any(CORBA::Any * any, CORBA::TypeCode_ptr tc)
{
#ifdef CLEAN_ORB_MICO

#if MICO_BIN_VERSION >= 0x020305
  any->set_type(tc);
#else
  any->type(tc);
#endif

#else

  // ORBacus
  any->replace(tc, 0);

#endif
}

////////////////////////////////
//      CStringArray
////////////////////////////////

typedef char ** CStringArray;
  
void 
c_string_array_create(int size, CStringArray * result)
{
  *result = new char*[size];
  assert(*result);
}

void 
c_string_array_set(CStringArray arr, int index, CleanString value)
			
{
  arr[index] = clean_strdup(value);
}

void
c_string_array_get(CStringArray arr, 
				   int index,
				   CleanString * out_value)
{
  *out_value = clean_string_create(arr[index]);
}

void
c_string_array_delete(CStringArray arr)
{
  delete[] arr;
}

////////////////////////////////
//          ORB
////////////////////////////////

void 
corba_orb_init(CStringArray argv, int argc, CORBA::ORB_ptr * result, 
			   CStringArray * out_argv, int * out_argc)
{
  *result = CORBA::ORB_init(argc, argv, "mico-local-orb");

  // Save for further use
  //
  // This makes it unnecessary to pass the orb as an argument to
  // functions which need to call ORB related functions, like
  // create_dyn_struct().
  //
  the_orb = *result;

  CORBA::Object_var obj =
    the_orb->resolve_initial_references ("DynAnyFactory");
  
  dynany_factory = DynamicAny::DynAnyFactory::_narrow (obj);
  assert(dynany_factory != (void*)NULL);

  obj = the_orb->resolve_initial_references("RootPOA");
  the_poa = PortableServer::POA::_narrow(obj);
  assert(the_poa);

  out.set_orb(the_orb);
  out.set_fac(dynany_factory);

  *out_argv = argv;
  *out_argc = argc;
}

void 
corba_orb_string_to_object(CORBA::ORB_ptr orb, 
						   CleanString str, 
						   CORBA::Object_ptr * result)
{
  char * s = clean_strdup(str);

  // ORBacus does not like the newline at the end of an IOR...
  if ((strlen(s) > 0) && (s[strlen(s) - 1] == '\n'))
	s[strlen(s) - 1] = 0;

  try {
	*result = orb->string_to_object(s);
  }
  catch (const CORBA::Exception& ex) {
	cout << ex << endl;
	*result = NULL;
  }

  delete[] s;
}

void 
corba_orb_bind(CORBA::ORB_ptr orb, 
			   CleanString repoid, 
			   CleanString address,
			   CORBA::Object_ptr * out_obj)
{
#ifdef CLEAN_ORB_MICO
  char * repoid2 = clean_strdup(repoid);
  char * address2 = clean_strdup(address);

  *out_obj = orb->bind(repoid2, address2);
  if (CORBA::is_nil(*out_obj)) {
	*out_obj = NULL;
  }
  delete[] repoid2;
  delete[] address2;
#else
  // Bind is MICO specific
  *out_obj = NULL;
#endif
}

void 
corba_orb_object_to_string(CORBA::ORB_ptr orb,
						   CORBA::Object_ptr obj,
						   CleanString * out_str)
{
  char * tmp;

  try {
	tmp = orb->object_to_string(obj);
	*out_str = clean_string_create(tmp);
  }
  catch (...) {
	*out_str = clean_string_create("");
  }

  CORBA::string_free(tmp);
}

void
corba_orb_resolve_initial_references(CORBA::ORB_ptr orb,
									 CleanString name,
									 CORBA::Object_ptr * out_obj)
{
  char * tmp = clean_strdup(name);

  try {
	*out_obj = orb->resolve_initial_references(tmp);
  }
  catch (...) {
	*out_obj = NULL;
  }

  delete[] tmp;
}

////////////////////////////////
//          OBJECT
////////////////////////////////

void 
corba_object_create_request(CORBA::Object_ptr obj,
						 CleanString operation,
						 CORBA::Request_ptr * out_req)
{
  char * tmp = clean_strdup(operation);

  *out_req = obj->_request(tmp);

  delete[] tmp;
}

void
corba_object_is_a(CORBA::Object_ptr obj,
				  CleanString repoid,
				  CORBA::Boolean * out_is_a)
{
  char * tmp = clean_strdup(repoid);

  try {
	*out_is_a = obj->_is_a(tmp);
  }
  catch (...) {
	*out_is_a = false;
  }
  delete[] tmp;
}

void
corba_object_nil(CORBA::Object_ptr * out_obj)
{
  *out_obj = CORBA::Object::_nil();
}

void
corba_object_hash(CORBA::Object_ptr obj,
				  CORBA::ULong max,
				  int * out_hash)
{
  *out_hash = obj->_hash(max);
}

////////////////////////////////
//         TYPECODE
////////////////////////////////

void
corba_struct_member_seq_create(int size,
							   CORBA::StructMemberSeq ** out_value)
{
  *out_value = new CORBA::StructMemberSeq();
  (*out_value)->length(size);
}

void
corba_struct_member_seq_fill(CORBA::StructMemberSeq * seq,
							 int index,
							 CleanString name,
							 CORBA::TypeCode_ptr type)
{
  char * tmp = clean_strdup(name);

  (*seq)[index].name = CORBA::string_dup(tmp);
  (*seq)[index].type = type;
  (*seq)[index].type_def = CORBA::IDLType::_nil();

  delete[] tmp;
}

void
corba_struct_member_seq_delete(CORBA::StructMemberSeq * seq)
{
  delete seq;
}

void
corba_union_member_seq_create(int size,
							  CORBA::UnionMemberSeq ** out_value)
{
  *out_value = new CORBA::UnionMemberSeq();
  (*out_value)->length(size);
}

void
corba_union_member_seq_fill(CORBA::UnionMemberSeq * seq,
							int index,
							CleanString name,
							CORBA::Any * label,
							CORBA::TypeCode_ptr type)
{
  char * tmp = clean_strdup(name);

  (*seq)[index].name = CORBA::string_dup(tmp);
  (*seq)[index].label = *label;
  (*seq)[index].type = type;

  delete[] tmp;
}

void
corba_union_member_seq_delete(CORBA::UnionMemberSeq * seq)
{
  delete seq;
}

void
corba_enum_member_seq_create(int size,
							 CORBA::EnumMemberSeq ** out_value)
{
  *out_value = new CORBA::EnumMemberSeq();
  (*out_value)->length(size);
}

void
corba_enum_member_seq_fill(CORBA::EnumMemberSeq * seq,
						   int index,
						   CleanString name)
{
  char * tmp = clean_strdup(name);

  (*seq)[index] = CORBA::string_dup(tmp);

  delete[] tmp;
}

void
corba_enum_member_seq_delete(CORBA::EnumMemberSeq * seq)
{
  delete seq;
}

void
corba_tc_basic(CORBA::TCKind kind, CORBA::TypeCode_ptr * out_tc)
{
#ifdef CLEAN_ORB_ORBACUS
  CORBA::TypeCode * tmp_tc;

  // ORBacus does not have a simple typecode creation function :(
  switch (kind) {
  case CORBA::tk_null:
	tmp_tc = CORBA::_tc_null;
	break;
  case CORBA::tk_void:
	tmp_tc = CORBA::_tc_void;
	break;

  case CORBA::tk_short: tmp_tc = CORBA::_tc_short; break;
  case CORBA::tk_long: tmp_tc = CORBA::_tc_long; break;
  case CORBA::tk_longlong: tmp_tc = CORBA::_tc_longlong; break;
  case CORBA::tk_ushort: tmp_tc = CORBA::_tc_ushort; break;
  case CORBA::tk_ulong: tmp_tc = CORBA::_tc_ulong; break;
  case CORBA::tk_ulonglong: tmp_tc = CORBA::_tc_ulonglong; break;
  case CORBA::tk_float: tmp_tc = CORBA::_tc_float; break;
  case CORBA::tk_double: tmp_tc = CORBA::_tc_double; break;
  case CORBA::tk_longdouble: tmp_tc = CORBA::_tc_longdouble; break;
  case CORBA::tk_boolean: tmp_tc = CORBA::_tc_boolean; break;
  case CORBA::tk_char: tmp_tc = CORBA::_tc_char; break;
  case CORBA::tk_wchar: tmp_tc = CORBA::_tc_wchar; break;
  case CORBA::tk_octet: tmp_tc = CORBA::_tc_octet; break;
  case CORBA::tk_any: tmp_tc = CORBA::_tc_any; break;
  case CORBA::tk_TypeCode: tmp_tc = CORBA::_tc_TypeCode; break;
  case CORBA::tk_Principal: tmp_tc = CORBA::_tc_Principal; break;
  default:
	assert(0);
  }

  *out_tc = CORBA::TypeCode::_duplicate(tmp_tc);
#else
  try {
	*out_tc = CORBA::TypeCode::create_basic_tc(kind);
  } 
  catch (...) {
	cout << "Error: create_tc_basic(" << (int)kind << ") failed." << endl;
	assert(0);
  }
#endif
}

void
corba_tc_string(int bound, CORBA::TypeCode_ptr * out_tc)
{
  try {
	*out_tc = the_orb->create_string_tc(bound);
  }
  catch (...) {
	cout << "Error: create_tc_string(" << bound << ") failed." << endl;
	assert(0);
  }
}

void
corba_tc_objref(CleanString repoid, 
				CleanString name,
				CORBA::TypeCode_ptr * out_tc)
{
  char * tmp_repoid = clean_strdup(repoid);
  char * tmp_name = clean_strdup(name);

  try {
	*out_tc = the_orb->create_interface_tc(tmp_repoid, tmp_name);
  }
  catch (...) {
	cout << "Error: create_interface_tc(" << tmp_repoid << ", " << tmp_name << ") failed." << endl;
	assert(0);
  }	

  delete[] tmp_repoid;
  delete[] tmp_name;
}

void
corba_tc_struct(CleanString repoid, 
				CleanString name,
				CORBA::StructMemberSeq * members,
				CORBA::TypeCode_ptr * out_tc)
{
  char * tmp_repoid = clean_strdup(repoid);
  char * tmp_name = clean_strdup(name);

  try {
	*out_tc = the_orb->create_struct_tc(tmp_repoid, tmp_name, *members);
  }
  catch (...) {
	cout << "Error: create_struct_tc(" << tmp_repoid << ", " << tmp_name << ") failed." << endl;
	for (unsigned int i = 0; i < members->length(); ++i) {
	  cout << "Member " << i << ": " << (*members)[i].name << endl;
	  out.display_IDL_TypeCode((*members)[i].type);
	}
	assert(0);
  }	

  delete[] tmp_repoid;
  delete[] tmp_name;
}

void
corba_tc_union(CleanString repoid, 
			   CleanString name,
			   CORBA::TypeCode_ptr label_type,
			   CORBA::UnionMemberSeq * members,
			   CORBA::TypeCode_ptr * out_tc)
{
  char * tmp_repoid = clean_strdup(repoid);
  char * tmp_name = clean_strdup(name);

  try {
	*out_tc = the_orb->create_union_tc(tmp_repoid, tmp_name, label_type, *members);
  }
  catch (...) {
	cout << "Error: create_union_tc(" << tmp_repoid << ", " << tmp_name << ") failed." << endl;
	cout << "Label type: " << endl;
	out.display_IDL_TypeCode(label_type);
	assert(0);
  }

  delete[] tmp_repoid;
  delete[] tmp_name;
}

void
corba_tc_sequence(int bound,
				  CORBA::TypeCode_ptr content_type,
				  CORBA::TypeCode_ptr * out_tc)
{
  try {
	*out_tc = the_orb->create_sequence_tc(bound, content_type);
  }
  catch (...) {
	cout << "Error: create_sequence_tc() failed." << endl;
	cout << "Content type: " << endl;
	out.display_IDL_TypeCode(content_type);
	assert(0);
  }
}

void
corba_tc_enum(CleanString repoid, 
			  CleanString name,
			  CORBA::EnumMemberSeq * members,
			  CORBA::TypeCode_ptr * out_tc)
{
  char * tmp_repoid = clean_strdup(repoid);
  char * tmp_name = clean_strdup(name);

  try {
	*out_tc = the_orb->create_enum_tc(tmp_repoid, tmp_name, *members);
  }
  catch (...) {
	cout << "Error: create_enum_tc() failed." << endl;
	assert(0);
  }

  delete[] tmp_repoid;
  delete[] tmp_name;
}

void
corba_tc_alias(CleanString repoid, 
			   CleanString name,
			   CORBA::TypeCode_ptr content_type,
			   CORBA::TypeCode_ptr * out_tc)
{
  char * tmp_repoid = clean_strdup(repoid);
  char * tmp_name = clean_strdup(name);

  try {
	*out_tc = the_orb->create_alias_tc(tmp_repoid, tmp_name, content_type);
  }
  catch (...) {
	cout << "Error: create_alias_tc() failed." << endl;
	assert(0);
  }

  delete[] tmp_repoid;
  delete[] tmp_name;
}

void
corba_tc_array(CORBA::TypeCode_ptr content_type,
			   int length,
			   CORBA::TypeCode_ptr * out_tc)
{
  try {
	*out_tc = the_orb->create_array_tc(length, content_type);
  }
  catch (...) {
	cout << "Error: create_array_tc() failed." << endl;
	assert(0);
  }
}

void
corba_tc_exception(CleanString repoid, 
				   CleanString name,
				   CORBA::StructMemberSeq * members,
				   CORBA::TypeCode_ptr * out_tc)
{
  char * tmp_repoid = clean_strdup(repoid);
  char * tmp_name = clean_strdup(name);

  try {
	*out_tc = the_orb->create_exception_tc(tmp_repoid, tmp_name, *members);
  }
  catch (...) {
	cout << "Error: create_exception_tc() failed." << endl;
	assert(0);
  }

  delete[] tmp_repoid;
  delete[] tmp_name;
}

void
corba_tc_recursive(CleanString repoid, 
				   CORBA::TypeCode_ptr * out_tc)
{
  char * tmp_repoid = clean_strdup(repoid);

  try {
	*out_tc = the_orb->create_recursive_tc(tmp_repoid);
  }
  catch (...) {
	cout << "Error: create_recursive_tc() failed." << endl;
	assert(0);
  }

  delete[] tmp_repoid;
}

void
corba_tc_release(CORBA::TypeCode_ptr tc)
{
  CORBA::release(tc);
}

// Accessor functions

void
corba_tc_get_kind(CORBA::TypeCode_ptr tc,
				  int * out_value)
{
  *out_value = tc->kind();
}

void
corba_tc_get_id(CORBA::TypeCode_ptr tc,
		CleanString * out_value)
{
  *out_value = clean_string_create(tc->id());
}

void
corba_tc_get_name(CORBA::TypeCode_ptr tc,
		  CleanString * out_value)
{
  *out_value = clean_string_create(tc->name());
}

void
corba_tc_get_member_count(CORBA::TypeCode_ptr tc,
			  int * out_value)
{
  *out_value = tc->member_count();
}

void
corba_tc_get_member_name(CORBA::TypeCode_ptr tc,
			 int index,
			 CleanString * out_value)
{
  *out_value = clean_string_create(tc->member_name(index));
}

void
corba_tc_get_member_type(CORBA::TypeCode_ptr tc,
			 int index,
			 CORBA::TypeCode_ptr * out_value)
{
  *out_value = tc->member_type(index);
}

void
corba_tc_get_member_label(CORBA::TypeCode_ptr tc,
			  int index,
			  CORBA::Any ** out_value)
{
  *out_value = tc->member_label(index);
}

void
corba_tc_get_discriminator_type(CORBA::TypeCode_ptr tc,
				CORBA::TypeCode_ptr * out_value)
{
  *out_value = tc->discriminator_type();
}

void
corba_tc_get_default_index(CORBA::TypeCode_ptr tc,
			   int * out_value)
{
  *out_value = tc->default_index();
}

void
corba_tc_get_length(CORBA::TypeCode_ptr tc,
		    int * out_value)
{
  *out_value = tc->length();
}

void
corba_tc_get_content_type(CORBA::TypeCode_ptr tc,
			  CORBA::TypeCode_ptr * out_value)
{
  *out_value = tc->content_type();
}

////////////////////////////////
//        ANY
////////////////////////////////

void 
corba_any_create_null(CORBA::Any ** out_any)
{
  // The default constructed Any has a type code of tc_null
  *out_any = new CORBA::Any;
}

void 
corba_any_create_short(int value,
		       CORBA::Any ** out_any)
{
  *out_any = new CORBA::Any;
  **out_any <<= static_cast<CORBA::Short>(value);
}

void 
corba_any_create_ushort(int value,
			CORBA::Any ** out_any)
{
  *out_any = new CORBA::Any;
  **out_any <<= static_cast<CORBA::UShort>(value);
}

void 
corba_any_create_long(int value,
		      CORBA::Any ** out_any)
{
  *out_any = new CORBA::Any;
  **out_any <<= static_cast<CORBA::Long>(value);
}

void 
corba_any_create_ulong(int value,
		       CORBA::Any ** out_any)
{
  *out_any = new CORBA::Any;
  **out_any <<= static_cast<CORBA::ULong>(value);
}

void 
corba_any_create_float(double value,
		       CORBA::Any ** out_any)
{
  *out_any = new CORBA::Any;
  **out_any <<= static_cast<CORBA::Float>(value);
}

void 
corba_any_create_double(double value,
			CORBA::Any ** out_any)
{
  *out_any = new CORBA::Any;
  **out_any <<= static_cast<CORBA::Double>(value);
}

void 
corba_any_create_bool(int value,
		      CORBA::Any ** out_any)
{
  *out_any = new CORBA::Any;
  **out_any <<= CORBA::Any::from_boolean(static_cast<CORBA::Boolean>(value));
}

void 
corba_any_create_char(int value,
					  CORBA::Any ** out_any)
{
  *out_any = new CORBA::Any;
  **out_any <<= CORBA::Any::from_char(static_cast<CORBA::Char>(value));
}

void 
corba_any_create_octet(int value,
					   CORBA::Any ** out_any)
{
  *out_any = new CORBA::Any;
  **out_any <<= CORBA::Any::from_octet(static_cast<CORBA::Octet>(value));
}

void 
corba_any_create_string(CleanString value,
						CORBA::Any ** out_any)
{
  char * tmp = clean_strdup(value);

  *out_any = new CORBA::Any;
  **out_any <<= tmp;

  delete[] tmp;
}

void 
corba_any_create_any(CORBA::Any * value,
					 CORBA::Any ** out_any)
{
  *out_any = new CORBA::Any;
  **out_any <<= value;
}

void 
corba_any_create_tc(CORBA::TypeCode_ptr value,
					CORBA::Any ** out_any)
{
  *out_any = new CORBA::Any;
  **out_any <<= value;
}

void 
corba_any_create_objref(CORBA::Object_ptr value,
			CORBA::Any ** out_any)
{
  *out_any = new CORBA::Any;

  **out_any <<= value;

  // MICO Does not provide <<= for Object_ptr ???
  //  **out_any <<= CORBA::Any::from_object(value, "Object");
}

void
corba_any_delete(CORBA::Any * any)
{
  delete any;
}


void
corba_any_get_type(CORBA::Any * any,
		   CORBA::TypeCode_ptr * out_tc)
{

  *out_tc = any->type();
}

void 
corba_any_extract_short(CORBA::Any * any,
			int * out_short)
{
  CORBA::Short tmp;

  assert( *(any) >>= tmp );

  *out_short = tmp;
}

void 
corba_any_extract_long(CORBA::Any * any,
		       int * out_long)
{
  CORBA::Long tmp;

  assert( *(any) >>= tmp );
  *out_long = tmp;
}

void 
corba_any_extract_ushort(CORBA::Any * any,
			 int * out_ushort)
{
  CORBA::UShort tmp;

  assert( *(any) >>= tmp );
  *out_ushort = tmp;
}

void 
corba_any_extract_ulong(CORBA::Any * any,
			int * out_ulong)
{
  CORBA::ULong tmp;

  assert( *(any) >>= tmp );
  *out_ulong = tmp;
}

void 
corba_any_extract_float(CORBA::Any * any,
			double * out_float)
{
  CORBA::Float tmp;

  assert( *(any) >>= tmp );
  *out_float = tmp;
}

void 
corba_any_extract_double(CORBA::Any * any,
			 double * out_double)
{
  CORBA::Double tmp;

  assert( *(any) >>= tmp );
  *out_double = tmp;
}

void 
corba_any_extract_boolean(CORBA::Any * any,
			  int * out_bool)
{
  CORBA::Boolean tmp;

  assert( *(any) >>= CORBA::Any::to_boolean(tmp) );
  *out_bool = tmp;
}

void 
corba_any_extract_char(CORBA::Any * any,
		       char * out_char)
{
  CORBA::Char tmp;

  assert( *(any) >>= CORBA::Any::to_char(tmp) );
  *out_char = tmp;
}

void 
corba_any_extract_octet(CORBA::Any * any,
			int * out_octet)
{
  CORBA::Octet tmp;

  assert( *(any) >>= CORBA::Any::to_octet(tmp) );
  *out_octet = tmp;
}

void
corba_any_extract_string(CORBA::Any * any,
			 CleanString * out_string)
{
#ifdef NON_CONST_ANY_EXTRACTION
  char * tmp;
#else
  const char * tmp;
#endif

  // The extraction operator
  assert( *(any) >>= tmp );
  *out_string = clean_string_create(tmp);
}

void
corba_any_extract_any(CORBA::Any * any,
		      const CORBA::Any ** out_value)
{
#ifdef NON_CONST_ANY_EXTRACTION
  CORBA::Any * value;
#else
  const CORBA::Any * value;
#endif

  assert( *(any) >>= value );
  *out_value = value;
}

void
corba_any_extract_tc(CORBA::Any * any,
		     CORBA::TypeCode_ptr * out_tc)
{
  CORBA::TypeCode_ptr tc;

  assert( *(any) >>= tc );
  *out_tc = tc;
}

void
corba_any_extract_objref(CORBA::Any * any,
			 CORBA::Object_ptr * out_value)
{
  CORBA::Object_ptr value;

  assert( *(any) >>= CORBA::Any::to_object(value) );
  *out_value = CORBA::Object::_duplicate(value);
}

////////////////////////////////
//        DYNANY
////////////////////////////////

void
corba_nv_pair_seq_create(int size,
						 DynamicAny::NameValuePairSeq ** out_value)
{
  *out_value = new DynamicAny::NameValuePairSeq();
  (*out_value)->length(size);
}

void
corba_nv_pair_seq_fill(DynamicAny::NameValuePairSeq * seq,
					   int index,
					   CleanString name,
					   CORBA::Any * value)
{
  char * tmp = clean_strdup(name);

  (*seq)[index].id = CORBA::string_dup(tmp);
  (*seq)[index].value = *value;

  delete[] tmp;
}

void
corba_nv_pair_seq_length(DynamicAny::NameValuePairSeq * seq,
			 int * out_length)
{
  *out_length = seq->length();
}

void
corba_nv_pair_seq_get(DynamicAny::NameValuePairSeq * seq,
					  int index,
					  CleanString * out_name,
					  CORBA::Any ** out_value)
{
  *out_name = clean_string_create((*seq)[index].id);
  *out_value = &((*seq)[index].value);
}

void
corba_nv_pair_seq_delete(DynamicAny::NameValuePairSeq * seq)
{
  delete seq;
}

void
corba_dyn_struct_create(CORBA::TypeCode_ptr tc,
						DynamicAny::NameValuePairSeq * members,
						CORBA::Any ** out_any)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any_from_type_code (tc);
  DynamicAny::DynStruct_var ds =
    DynamicAny::DynStruct::_narrow (tmp);
  assert(!CORBA::is_nil(ds));

  /* FOR TESTING:
  DynamicAny::NameValuePairSeq  * components = ds->get_members();

  for (unsigned int i = 0; i < members->length(); ++i) {
	if (! ((*members)[i].value.type())->equivalent((*components)[i].value.type())) {
	  cout << (*members)[i].value.type()->id() << endl;
	  cout << (*components)[i].value.type()->id() << endl;
	  //	  out.display_IDL_TypeCode(((*members)[i]).value.type());
	  //	  out.display_IDL_TypeCode(((*components)[i]).value.type());
	}
  }
  */

  try {
	ds->set_members(*members);
  }
  catch (...) {
	cout << "Error: DynStruct::set_members() failed." << endl;
	assert(0);
  }

  *out_any = ds->to_any();

  ds->destroy();
}

void
corba_dyn_struct_members(CORBA::Any * any,
						 DynamicAny::NameValuePairSeq ** out_members)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any(*any);
  DynamicAny::DynStruct_var ds =
    DynamicAny::DynStruct::_narrow (tmp);
  assert(!CORBA::is_nil(ds));

  *out_members = ds->get_members();

  ds->destroy();
}

void
corba_any_seq_create(int size,
					 DynamicAny::AnySeq ** out_value)
{
  *out_value = new DynamicAny::AnySeq();
  (*out_value)->length(size);
}

void
corba_any_seq_fill(DynamicAny::AnySeq * seq,
				   int index,
				   CORBA::Any * value)
{
  (*seq)[index] = *value;
}

void
corba_any_seq_length(DynamicAny::AnySeq * seq,
					 int * out_length)
{
  *out_length = seq->length();
}

void
corba_any_seq_get(DynamicAny::AnySeq * seq,
				  int index,
				  CORBA::Any ** out_value)
{
  *out_value = &((*seq)[index]);
}

void
corba_any_seq_delete(DynamicAny::AnySeq * seq)
{
  delete seq;
}

void
corba_dyn_sequence_create(CORBA::TypeCode_ptr tc,
						  DynamicAny::AnySeq * elements,
						  CORBA::Any ** out_any)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any_from_type_code (tc);
  DynamicAny::DynSequence_var ds =
    DynamicAny::DynSequence::_narrow (tmp);
  assert(!CORBA::is_nil(ds));

  ds->set_length(elements->length());
  ds->set_elements(*elements);
  *out_any = ds->to_any();

  ds->destroy();
}

void
corba_dyn_sequence_elements(CORBA::Any * any,
							DynamicAny::AnySeq ** out_elements)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any(*any);
  DynamicAny::DynSequence_var ds =
    DynamicAny::DynSequence::_narrow (tmp);
  assert(!CORBA::is_nil(ds));

  *out_elements = ds->get_elements();

  DynamicAny::AnySeq * seq = ds->get_elements();
  delete seq;

  ds->destroy();
}

void
corba_dyn_union_create(CORBA::TypeCode_ptr tc,
					   CORBA::Any * label,
					   CORBA::Any * value,
					   CORBA::Any ** out_any)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any_from_type_code(tc);
  DynamicAny::DynUnion_var du =
    DynamicAny::DynUnion::_narrow (tmp);
  assert(!CORBA::is_nil(du));

  du->set_discriminator(dynany_factory->create_dyn_any(*label));
  if (! (du->has_no_active_member())) {
	// ORBacus needs this...
	du->rewind();
	du->next();
	du->current_component()->from_any(*value);
  }
  *out_any = du->to_any();

  du->destroy();
}

void
corba_dyn_union_get(CORBA::Any * any,
					CORBA::Any ** out_label,
					CORBA::Any ** out_member)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any(*any);
  DynamicAny::DynUnion_var du =
    DynamicAny::DynUnion::_narrow (tmp);
  assert(!CORBA::is_nil(du));

  *out_label = du->get_discriminator()->to_any();
  if (du->has_no_active_member()) {
	// There must be a better way to create an Any...
	// This any has type code tk_null
	*out_member = new CORBA::Any;
  }
  else {
	*out_member = du->member()->to_any();
  }

  du->destroy();
}

void
corba_dyn_enum_create(CORBA::TypeCode_ptr tc,
					  CleanString value,
					  CORBA::Any ** out_any)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any_from_type_code(tc);
  DynamicAny::DynEnum_var de =
    DynamicAny::DynEnum::_narrow (tmp);
  assert(!CORBA::is_nil(de));

  de->set_as_string(clean_strdup(value));

  *out_any = de->to_any();

  de->destroy();
}

void
corba_dyn_enum_get(CORBA::Any * any,
				   CleanString * out_value)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any(*any);
  DynamicAny::DynEnum_var de =
    DynamicAny::DynEnum::_narrow (tmp);
  assert(!CORBA::is_nil(de));

  *out_value = clean_string_create(de->get_as_string());

  de->destroy();
}

void
corba_dyn_array_create(CORBA::TypeCode_ptr tc,
					   DynamicAny::AnySeq * elements,
					   CORBA::Any ** out_any)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any_from_type_code (tc);
  DynamicAny::DynArray_var da =
    DynamicAny::DynArray::_narrow (tmp);
  assert(!CORBA::is_nil(da));

  da->set_elements(*elements);
  *out_any = da->to_any();

  da->destroy();
}

void
corba_dyn_array_elements(CORBA::Any * any,
						 DynamicAny::AnySeq ** out_elements)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any(*any);
  DynamicAny::DynArray_var da =
    DynamicAny::DynArray::_narrow (tmp);
  assert(!CORBA::is_nil(da));

  *out_elements = da->get_elements();

  da->destroy();
}

////////////////////////////////
//       NAMEDVALUE
////////////////////////////////

void
corba_nv_get_name(CORBA::NamedValue_ptr nv,
				  CleanString * out_name)
{
  *out_name = clean_string_create(nv->name());
}

void
corba_nv_get_value(CORBA::NamedValue_ptr nv,
				   CORBA::Any ** out_value)
{
  *out_value = nv->value();
}

void
corba_nv_get_flags(CORBA::NamedValue_ptr nv,
				   int * out_flags)
{
  *out_flags = nv->flags();
}

void
corba_nv_set_type(CORBA::NamedValue_ptr nv,
				  CORBA::TypeCode_ptr tc)
{
  set_type_of_any(nv->value(), tc);
}

void
corba_nv_set_value(CORBA::NamedValue_ptr nv,
				   CORBA::Any * value)
{
  (*nv->value()) = *value;
}

////////////////////////////////
//           NVLIST
////////////////////////////////

void
corba_nvlist_create(CORBA::NVList_ptr * out_nvlist)
{
  the_orb->create_list(0, *out_nvlist);
}

void
corba_nvlist_add(CORBA::NVList_ptr nvlist,
				 int flags,
				 CORBA::NamedValue_ptr * out_nv)
{
  *out_nv = nvlist->add(flags);
}

void
corba_nvlist_count(CORBA::NVList_ptr nvlist,
				   int * out_count)
{
  *out_count = nvlist->count();
}

void
corba_nvlist_get(CORBA::NVList_ptr nvlist,
				 int index,
				 CORBA::NamedValue_ptr * out_nv)
{
  *out_nv = nvlist->item(index);
}

////////////////////////////////
//       REQUEST
////////////////////////////////

void
corba_request_arguments(CORBA::Request_ptr req,
						CORBA::NVList_ptr * out_nvlist)
{
  *out_nvlist = req->arguments();
}

void 
corba_request_add_in_arg(CORBA::Request_ptr req,
						 CleanString name,
						 CORBA::Any * value)
{
  char * tmp = clean_strdup(name);

  (req->add_in_arg(tmp)) = *value;
  delete[] tmp;
}

void 
corba_request_add_out_arg(CORBA::Request_ptr req,
			  CleanString name,
			  CORBA::Any * value)
{
  char * tmp = clean_strdup(name);

  (req->add_out_arg(tmp)) = *value;
  delete[] tmp;
}

void 
corba_request_add_inout_arg(CORBA::Request_ptr req,
			    CleanString name,
			    CORBA::Any * value)
{
  char * tmp = clean_strdup(name);

  (req->add_inout_arg(tmp)) = *value;
  delete[] tmp;
}

void 
corba_request_add_out_arg_template(CORBA::Request_ptr req,
								   CleanString name,
								   CORBA::TypeCode_ptr tc)
{
  char * tmp = clean_strdup(name);

  set_type_of_any(&(req->add_out_arg(tmp)), tc);

  delete[] tmp;
}

void 
corba_request_invoke(CORBA::Request_ptr req)
{
#ifdef CLEAN_ORB_MICO
  // Apparently, if a SystemException occurs during the invoke() call,
  // MICO raises it as a C++ exception instead of putting it into the
  // environment.
  // TODO: Is this still true in 2.3.5?
  try {
	req->invoke();
  }
  catch (const CORBA::SystemException& e) {
	req->env()->exception(new CORBA::SystemException(e));
  }
#else
	req->invoke();
#endif
}

void 
corba_request_set_result_tc(CORBA::Request_ptr req,
							CORBA::TypeCode_ptr tc)
{
  set_type_of_any(req->result()->value(), tc);
}
  
void 
corba_request_result(CORBA::Request_ptr req,
					 CORBA::Any ** out_any)
{
  *out_any = req->result()->value();
}

void
corba_request_exception(CORBA::Request_ptr req,
						CORBA::Exception ** out_ex)
{
  *out_ex = req->env()->exception();
}

void
corba_request_add_exception_def(CORBA::Request_ptr req,
								CORBA::TypeCode_ptr tc)
{
  //  out.display_IDL_TypeCode(tc);
  req->exceptions()->add(tc);
}

void
corba_request_delete(CORBA::Request_ptr req)
{
#ifdef CLEAN_ORB_MICO
  delete req;
#else
  release(req);
#endif
}

////////////////////////////////
//    EXCEPTION
////////////////////////////////

void
corba_exception_is_system(CORBA::Exception * ex,
						  int * out_bool)
{
  // For some reason, dynamic_cast<> crashes on NT with VC6 and MICO 2.3.5 :(
  *out_bool = (CORBA::SystemException::_downcast(ex) != 0);
}

void
corba_exception_get_minor(CORBA::Exception * ex,
						  int * out_minor)
{
  CORBA::SystemException * se = CORBA::SystemException::_downcast(ex);
  assert(se != NULL);
  *out_minor = se->minor();
}

void
corba_exception_get_completed(CORBA::Exception * ex,
							  int * out_completed)
{
  CORBA::SystemException * se = CORBA::SystemException::_downcast(ex);
  assert(se != NULL);
  *out_completed = se->completed();
}

void
corba_exception_get_repoid(CORBA::Exception * ex,
						   CleanString * out_repoid)
{
#ifdef CLEAN_ORB_MICO
  *out_repoid = clean_string_create(ex->_repoid());
#else
  *out_repoid = clean_string_create(ex->_rep_id());
#endif
}

void
corba_unknown_ex_narrow(CORBA::Exception * ex,
						CORBA::UnknownUserException ** out_ex)
{
  *out_ex = CORBA::UnknownUserException::_downcast(ex);
}

// Answer the repoid of the user exception contained in the exception
void
corba_unknown_ex_get_repoid(CORBA::UnknownUserException * ex,
							CleanString * out_repoid)
{
  // TODO: Fix memory leaks in this code
  *out_repoid = clean_string_create(ex->exception().type()->id());
}

// Answer the user exception contained in the exception
void
corba_unknown_ex_get_value(CORBA::UnknownUserException * ex,
						   CORBA::TypeCode_ptr tc,
						   CORBA::Any ** out_any)
{
  //
  // This is not documented in the CORBA specification.
  // The 'exception(tc)' call tries to unmarshall the Any
  // contained in Exception using the TypeCode provided by the user.
  // If it succeeds, it returns an Any. Otherwise, it asserts....
  //

#ifdef CLEAN_MICO_ORB
  // Does the TypeCode argument necessary ?
  *out_any = &(ex->exception(tc));
#else
  *out_any = &(ex->exception());
#endif
}

////////////////////////////////
//      SERVER REQUEST
////////////////////////////////

void
corba_server_request_operation(CORBA::ServerRequest_ptr req,
							   CleanString * out_str)
{
  *out_str = clean_string_create(req->operation());
}

void
corba_server_request_set_params(CORBA::ServerRequest_ptr req,
								CORBA::NVList_ptr args)
{
  req->arguments(args);
}

void
corba_server_request_set_result(CORBA::ServerRequest_ptr req,
								CORBA::Any * result)
{
  req->set_result(*result);
}

void
corba_server_request_set_exception(CORBA::ServerRequest_ptr req,
								   CORBA::Any * result)
{
  req->set_exception(*result);
}

void
corba_server_request_get_args(CORBA::ServerRequest_ptr req,
							  CORBA::NVList_ptr * result)
{
#ifdef CLEAN_ORB_MICO
  // begin-mico-extension
  *result = req->arguments();
  // end-mico-extension
#else
  // TODO:
  assert(0);
#endif
}

////////////////////////////////
//       SERVER SIDE
////////////////////////////////

static pthread_t the_listener_thread;

// This signals to the Clean thread that a CORBA request is received
static pthread_mutex_t request_mutex;

static pthread_cond_t request_cond;

// This signals to the CORBA thread that the reply to the request has been
// computed
static pthread_mutex_t reply_mutex;

static pthread_cond_t reply_cond;

// The received request
static CORBA::ServerRequest_ptr the_request;

static CORBA::Object_ptr the_object;

// The ObjectId type does not have a comparison operator, thus it can not
// be stored in STL maps.
bool
operator< (const PortableServer::ObjectId& o1, const PortableServer::ObjectId & o2);

// Mapping ObjectIds to repository IDs
static map<PortableServer::ObjectId, char *, less<PortableServer::ObjectId> > repository_ids;

// The repository ID of the newly created CORBA Object
static char * the_repo_id;

class MyServant : virtual public PortableServer::DynamicImplementation {
  
  virtual void invoke (CORBA::ServerRequest_ptr req) {
	pthread_mutex_lock(&reply_mutex);

	the_request = req;
	the_object 	= _this();

	// TODO: Wait until the Clean thread is ready to receive the request
	pthread_mutex_lock(&request_mutex);
	pthread_cond_signal(&request_cond);
	pthread_mutex_unlock(&request_mutex);

	pthread_cond_wait(&reply_cond, &reply_mutex);

	pthread_mutex_unlock(&reply_mutex);
  }

  virtual CORBA::RepositoryId _primary_interface(const PortableServer::ObjectId& id, PortableServer::POA_ptr poa) {
	// According to the CORBA 2.2 spec, the ORB only calls this function
	// when serving a request. However, under MICO 2.3.0, the ORB calls
	// this function when the application calls 
	// PortableServer::activate_object()...

	// Since the primary_interface of each object is permanent, we
	// store it in an STL map
	if (repository_ids.find(id) == repository_ids.end()) {
	  repository_ids[id] = strdup(the_repo_id);
	  //	  delete[] the_repo_id;
	}

    return(CORBA::string_dup(repository_ids[id]));
  }
};

bool
operator< (const PortableServer::ObjectId& o1, const PortableServer::ObjectId & o2)
{
  if (o1.length() != o2.length()) {
    return (o1.length() < o2.length());
  }

  for (unsigned int i = 0; i < o1.length(); ++i)
	if (o1[i] < o2[i])
	  return(true);
  return(false);
}

void *
corba_listener_thread_main(void * arg)
{
  //  cout << "Listener thread started." << endl;

  PortableServer::POAManager_var mgr = the_poa->the_POAManager();
  mgr->activate();

  the_orb->run();

  return(NULL);
}

void
corba_listener_init()
{
  int err;

  pthread_mutex_init(&request_mutex, NULL);
  pthread_mutex_init(&reply_mutex, NULL);

  pthread_cond_init(&request_cond, NULL);
  pthread_cond_init(&reply_cond, NULL);

  err = pthread_create(&the_listener_thread, NULL, 
					   corba_listener_thread_main, NULL);
  assert(err == 0);
  pthread_detach(the_listener_thread);
}

void
corba_listener_receive(CORBA::ServerRequest_ptr * out_req,
					   CORBA::Object_ptr * out_obj)
{
  //  cout << "Receiving..." << endl;

  // Wait for a request
  pthread_mutex_lock(&request_mutex);
  pthread_cond_wait(&request_cond, &request_mutex);
  pthread_mutex_unlock(&request_mutex);  

  //  cout << "Request received." << endl;

  // return request to the Clean level
  *out_req = the_request;
  *out_obj = the_object;
}

void
corba_listener_reply()
{
  // Signal the CORBA thread that the reply has been computed
  pthread_mutex_lock(&reply_mutex);
  pthread_cond_signal(&reply_cond);
  pthread_mutex_unlock(&reply_mutex);  

  //  cout << "Reply sent..." << endl;
}

void
corba_servant_create(CleanString repoid,
					 CORBA::Object_ptr * out_obj)
{
  // This is freed in MyServant::_primary_interface()
  the_repo_id = clean_strdup(repoid);

  PortableServer::ObjectId * id = the_poa->activate_object(new MyServant);  
  *out_obj = the_poa->id_to_reference(*id);
}


} // extern "C"  

  

// Local Variables:
// tab-width: 4
// End:



