module ir

import StdEnv, ArgEnv
import Corba, repository

import RWSDebug, Debug

////////////////////////////////
//       UTILITIES
////////////////////////////////

no_rule_match :: String a -> .b
no_rule_match fct_name value = abort ("\n\nError: no pattern matches " +++ (toStr value) +++ " for function " +++ fct_name +++ "\n\n")

with_indexes :: [a] -> [(a, Int)]
with_indexes l = zip2 l [0..]

//
//   HIGHER LEVEL INTERFACE TO THE INTERFACE REPOSITORY
//

:: IRObject = IRRepository [IRObject]
			| IRModule CORBA_ModuleDescription [IRObject]
			| IRConstant CORBA_ConstantDescription
			| IRStruct CORBA_TypeDescription [IRObject]
			| IRException CORBA_ExceptionDescription
			| IRUnion CORBA_TypeDescription [IRObject]
			| IREnum CORBA_TypeDescription
			| IRAlias CORBA_TypeDescription
			| IRInterface CORBA_InterfaceDescription [IRObject]
			| IROperation CORBA_OperationDescription
			| IRAttribute CORBA_AttributeDescription

ir_dump :: CORBA_IRObject *World -> (IRObject, *World)
ir_dump obj env
	# (res, env) = CORBA_IRObject__get_def_kind obj env
	# (Result def_kind) = res
	= dump_by_type obj def_kind env
	where
		dump_by_type :: CORBA_IRObject CORBA_DefinitionKind *World -> (IRObject, *World)

		dump_by_type obj Dk_Interface env
		# (desc, env) = get_description obj env
		# (Just desc, env) = CORBA_InterfaceDescription__decode desc env
		# (members, env) = get_contents obj env
		= (IRInterface desc members, env)

		dump_by_type obj Dk_Module env
		# (desc, env) = get_description obj env
		# (Just desc, env) = CORBA_ModuleDescription__decode desc env
		# (members, env) = get_contents obj env
		= (IRModule desc members, env)

		dump_by_type obj Dk_Repository env
		# (members, env) = get_contents obj env 
		= (IRRepository members, env)

		dump_by_type obj Dk_Struct env
		# (desc, env) = get_description obj env
		# (Just desc, env) = CORBA_TypeDescription__decode desc env
		# (members, env) = get_contents obj env
		= (IRStruct desc members, env)

		dump_by_type obj Dk_Union env
		# (desc, env) = get_description obj env
		# (Just desc, env) = CORBA_TypeDescription__decode desc env
		# (members, env) = get_contents obj env 
		= (IRUnion desc members, env)

		dump_by_type obj Dk_Enum env
		# (desc, env) = get_description obj env
		# (Just desc, env) = CORBA_TypeDescription__decode desc env
		= (IREnum desc, env)

		dump_by_type obj Dk_Alias env
		# (desc, env) = get_description obj env
		# (Just desc, env) = CORBA_TypeDescription__decode desc env
		= (IRAlias desc, env)

		dump_by_type obj Dk_Exception env
		# (desc, env) = get_description obj env
		# (Just desc, env) = CORBA_ExceptionDescription__decode desc env
		= (IRException desc, env)

		dump_by_type obj Dk_Operation env
		# (desc, env) = get_description obj env
		# (Just desc, env) = CORBA_OperationDescription__decode desc env
		= (IROperation desc, env)

		dump_by_type obj Dk_Attribute env
		# (desc, env) = get_description obj env
		# (Just desc, env) = CORBA_AttributeDescription__decode desc env
		= (IRAttribute desc, env)

		dump_by_type obj Dk_Constant env
		# (desc, env) = get_description obj env
		# (Just desc, env) = CORBA_ConstantDescription__decode desc env
		= (IRConstant desc, env)

		dump_by_type obj kind env = no_rule_match "dump_by_type" kind env

		get_description :: CORBA_IRObject *World -> (Any, *World)
		get_description obj env
		# (obj, env) = CORBA_Contained__narrow obj env
		# (Result desc, env) = CORBA_Contained_describe (just obj) env
		= (desc.CORBA_Contained_Description.value, env)

		get_contents :: CORBA_IRObject *World -> ([IRObject], *World)
		get_contents obj env
		# (obj, env) = CORBA_Container__narrow obj env
		# (Result seq, env) = CORBA_Container_contents (just (obj)) Dk_all True env
		# (res, env) = map_with_env dump_member seq env
		= (res, env)
		where
			dump_member :: CORBA_Contained *World -> (IRObject, *World)
			dump_member obj env
			# (member, env) = ir_dump (CORBA_IRObject__widen obj) env
			= (member, env)


/////////////////////////////////
//     UTILITY FUNCTIONS
/////////////////////////////////

string_split :: String Char -> [String]
string_split input c = f (fromString input) [] []
where
	f :: [Char]	[String] [Char] -> [String]
	f [] result part = result ++ [(toString part)]
	f [x:xs] result part
		| x == c
		= f xs (result ++ [(toString part)]) []
		| otherwise
		= f xs result (part ++ [x])

string_replace_char :: String Char Char -> String
string_replace_char input from_c to_c = toString (map f (fromString input))
where
	f c = if (c == from_c) to_c c

string_capitalize_first_char :: String -> String
string_capitalize_first_char input = toString (f (fromString input))
where
	f [] = []
	f [x:xs] = [toUpper x:xs]

// Generate the Clean name from the CORBA Repository ID
// It is assumed that the repository id is in OMG IDL format.
gen_clean_name :: String -> String
gen_clean_name repoid = (string_replace_char (remove_omg_prefix ((string_split repoid ':')!!1))) '/' '_'

// Remove the 'omg.org' prefix, it it exists
remove_omg_prefix :: String -> String
remove_omg_prefix repoid = toString (f (splitAt 8 (fromString repoid)))
where
	f (['o', 'm', 'g', '.', 'o', 'r', 'g', '/'], rest) = rest
	f (head, tail) = head ++ tail

// Convert the list into a string, inserting a delimiter between elements
toStringWith :: [a] String -> String | toString a
toStringWith [] _	=	""
toStringWith [x] _	=	toString x
toStringWith [x:xs]	delim =	(toString x) +++ delim +++ (toStringWith xs delim)

// Convert the list into a string, inserting a delimiter before elements
toStringWith2 :: [a] String -> String | toString a
toStringWith2 []     _		=	""
toStringWith2 [x]    delim	=	delim +++ (toString x)
toStringWith2 [x:xs] delim 	=	delim +++ (toString x) +++ (toStringWith2 xs delim)

stringify :: String -> String
stringify s = "\"" +++ s +++ "\""

charify :: Char -> String
charify c = "'" +++ (toString c) +++ "'"

concat_list :: [String] -> String
concat_list l = foldl (+++) "" l

// Replace the '%' symbols in arg1 with arg2
string_subst :: String String -> String
string_subst s1 s2 = foldl (+++) "" (map f (fromString s1))
where
	f '%'	=	s2
	f c		=	toString c

// Answer the Clean type name for the given TypeCode.
clean_type_for :: TypeCode -> String
clean_type_for tc = case tc of
 	Tk_Void			->	"CORBA_Void"
 	Tk_Short		->	"CORBA_Short"
 	Tk_Long			->	"CORBA_Long"
 	Tk_UShort		->	"CORBA_UShort"
	Tk_ULong		->	"CORBA_ULong"
	Tk_Float		->	"CORBA_Float"
	Tk_Double		->	"CORBA_Real"
	Tk_Boolean		->	"CORBA_Boolean"
	Tk_Char			->	"CORBA_Char"
	Tk_Octet		->	"CORBA_Octet"
	Tk_Any			->	"CORBA_Any"
	(Tk_String _)	->	"CORBA_String"
	Tk_TypeCode		->	"CORBA_TypeCode"
	(Tk_ObjRef repoid _)	->	gen_clean_name repoid
	(Tk_Struct repoid _ _) 	->	gen_clean_name repoid
	(Tk_Union repoid _ _ _)	->	gen_clean_name repoid
	(Tk_Enum repoid _ _)	->	gen_clean_name repoid
	(Tk_Alias repoid _ _)	->	gen_clean_name repoid
	(Tk_Exception repoid _ _)	->	gen_clean_name repoid
	(Tk_Recursive repoid)	->	gen_clean_name repoid
	tc				->	abort ("clean_type_for called with" +++ (toStr tc))

// Same as above, but handle anonymous types too
typename_for :: TypeCode String -> String
typename_for tc name
	= if (is_anonym_member tc) name (clean_type_for tc)

// Answer a Clean expression representing this TypeCode
// The second argument is used to name anonymous types
typecode_for :: TypeCode String -> String
typecode_for (Tk_String bound) _ 
	= "(Tk_String " +++ (toString bound) +++ ")"
typecode_for tc name 
	= (typename_for tc name) +++ "__tc"

// Answer the name of the decoder function of this type
decode_for :: TypeCode -> String
decode_for tc = (clean_type_for tc) +++ "__decode"

// Answer the name of the encoder function of this type
encode_for :: TypeCode -> String
encode_for tc = (clean_type_for tc) +++ "__encode"

// Answer the name of the function returning a default value of this type
// This is used in constructing OUT arguments
default_value_for :: TypeCode -> String
default_value_for tc = (clean_type_for tc) +++ "__default"

clean_type_for_anonym_member :: String String -> String
clean_type_for_anonym_member parent_name member_name = parent_name +++ "__members__" +++ member_name

// Answer whenever this typecode denotes an anonymous type
is_anonym_member :: TypeCode -> Bool
is_anonym_member (Tk_Sequence _ _) 	=	True
is_anonym_member (Tk_Array _ _) 	=	True
is_anonym_member _					=	False

clean_value_for any = case any of
	(CorbaShort i)	->	toString i
	(CorbaLong i)	->	toString i
	(CorbaUShort i)	->	toString i
	(CorbaULong i)	->	toString i
	(CorbaFloat f)	->	toString f
	(CorbaDouble f)	->	toString f
	(CorbaBoolean b)->	toString b
	(CorbaChar c)	->	charify c
	(CorbaOctet i)	->	toString i
	(CorbaString s)	->	stringify s
	(CorbaEnum tc=:(Tk_Enum repoid _ _) s)-> string_capitalize_first_char s

// Return a Clean expression representing this Any
clean_encoding_for any = case any of
	(CorbaShort i)	->	"(CorbaShort " +++ (toString i) +++ ")"
	(CorbaLong i)	->	"(CorbaLong " +++ (toString i) +++ ")"
	(CorbaUShort i)	->	"(CorbaUShort " +++ (toString i) +++ ")"
	(CorbaULong i)	->	"(CorbaULong " +++ (toString i) +++ ")"
	(CorbaFloat f)	->	"(CorbaFloat " +++ (toString f) +++ ")"
	(CorbaDouble f)	->	"(CorbaDouble " +++ (toString f) +++ ")"
	(CorbaBoolean b)->	"(CorbaBoolean " +++ (toString b) +++ ")"
	(CorbaChar c)	->	"(CorbaChar " +++ (toString c) +++ ")"
	(CorbaOctet i)	->	"(CorbaOctet " +++ (toString i) +++ ")"
	(CorbaEnum tc=:(Tk_Enum repoid _ _) s)-> "(CorbaEnum " +++ (typecode_for tc "") +++ " " +++ s +++ ")"

// Generate a header comment for the given definition
gen_header_comment :: String -> String
gen_header_comment name = "\n" +++ (toString (repeatn 32 '/')) +++ "\n//     " +++ name +++ "\n" +++ (toString (repeatn 32 '/')) +++ "\n"

gen_decode_def_for :: String -> String
gen_decode_def_for name	= name +++ "__decode :: Any *World -> (MayBe " +++ name +++ ", *World)\n"

// Answer the pair of get/set operations for the given attribute
get_operations_for :: CORBA_AttributeDescription CORBA_IDLType -> [CORBA_OperationDescription]
get_operations_for attr idltype_nil
	= [read_desc] ++ (if read_write [write_desc] [])
where
		read_desc = { CORBA_OperationDescription |
			name 		= "_get_" +++ attr.CORBA_AttributeDescription.name,
			id	 		= "foo",  // A bit ugly...
			defined_in 	= attr.CORBA_AttributeDescription.defined_in,
			version 	= "1.0",
			result 		= attr.CORBA_AttributeDescription.type,
			mode		= OP_NORMAL,
			contexts	= [],
			parameters	= [],
			exceptions	= []
			}

		read_write	= (attr.CORBA_AttributeDescription.mode == ATTR_NORMAL)

		write_desc = { CORBA_OperationDescription |
			name 		= "_set_" +++ attr.CORBA_AttributeDescription.name,
			id	 		= "foo",  // A bit ugly...
			defined_in 	= attr.CORBA_AttributeDescription.defined_in,
			version 	= "1.0",
			result 		= Tk_Void,
			mode		= OP_NORMAL,
			contexts	= [],
			parameters	= [write_arg],
			exceptions	= []
			}		

		write_arg = { CORBA_ParameterDescription |
			name		= "attr",
			type		= attr.CORBA_AttributeDescription.type,
			mode		= PARAM_IN,
			// This field contains an object reference, but I need a
			// *World to create one. Hence this hack...
			type_def	= idltype_nil
			}


////////////////////////////////
//      IDL COMPILER
////////////////////////////////

:: CompilerState  = { 	
	genDef 			:: Bool,
	// The 'compile_attribute' function
	// needs an object of this type, hence this hack...
	idlType_nil 	:: CORBA_IDLType,
	// Only used by compile_operation. Signals that server side
	// operation definitions should be generated
	genServantDef	:: Bool,
	// Only used by compile_operation. Signals that server side 
	// wrapper functions should be generated
	genWrapper		:: Bool
					 }

compile_ir :: CompilerState IRObject -> String
compile_ir opt ir = case ir of
	(IREnum desc)				->	compile_enum opt desc
	(IRStruct desc members)		->	compile_struct opt desc members
	(IRRepository members) 		->	compile_repository opt members
	(IRInterface desc members) 	->	compile_interface opt desc members
	(IRModule desc members) 	->	compile_module opt desc members
	(IRConstant desc) 			->	compile_constant opt desc
	(IRException desc) 			->	compile_exception opt desc
	(IRUnion desc members) 		->	compile_union opt desc members
	(IRAlias desc) 				->	compile_alias opt desc
	(IROperation desc) 			->	compile_operation opt desc
	(IRAttribute desc) 			->	compile_attribute opt desc

////////////////////////////////
//    ANONYM TYPES
////////////////////////////////

compile_anonym_type :: CompilerState String TypeCode -> String
compile_anonym_type opt name tc=:(Tk_Sequence _ _) 
	= compile_sequence opt name tc
compile_anonym_type opt name tc=:(Tk_Array _ _) 
	= compile_array opt name tc

////////////////////////////////
//        SEQUENCES
////////////////////////////////

compile_sequence :: CompilerState String TypeCode -> String
compile_sequence opt name (Tk_Sequence length tc)
	= foldl (+++) "" (map (\s -> (string_subst s clean_name)) components)
	where
		components = (if opt.genDef def_components impl_components)

		def_components
			= [gen_member_type,
			   gen_comment, gen_def, gen_tc_def, gen_decode_def, 
			   gen_encode_def]
		impl_components 
			= [gen_member_type, gen_comment, gen_def, gen_tc_def, 
			   gen_tc_impl, gen_decode_def, gen_decode_impl, 
			   gen_encode_def, gen_encode_impl]

		// Code generator functions
		// The '%' symbol is substituted for the name of the clean type

		gen_comment = gen_header_comment clean_name

		gen_member_type = if (is_anonym_member tc) (compile_anonym_type opt member_type_name tc) ""

		gen_unique = ""

		gen_def = "\n:: " +++ gen_unique +++ clean_name +++ " :== [" +++ (member_type_name) +++ "]\n"
		
		member_type_name = if (is_anonym_member tc) (clean_name +++ "__member_type") (clean_type_for tc)
		
		gen_tc_def = 
			clean_name +++ "__tc :: TypeCode\n"

		gen_tc_impl = 
			clean_name +++ "__tc = Tk_Sequence " +++ (toString length) +++ " (" +++ (typecode_for tc member_type_name) +++ ")\n"

		gen_decode_def	= gen_decode_def_for clean_name

		gen_decode_impl	= 
			"%__decode (CorbaSequence tc members) env\n" 					+++
			"\t| tc <> " +++ (typecode_for tc member_type_name) +++ "\n" 	+++
			"\t= (Nothing, env)\n" 											+++
			"\t| otherwise\n" 												+++
			"\t= map_maybe_with_env " +++ (member_type_name) +++ "__decode members env\n"																  +++
			"%__decode _ env = (Nothing, env)\n"
						  
		gen_encode_def	= "%__encode :: % -> Any\n"

		gen_encode_impl	= "%__encode value = (CorbaSequence %__tc (map " +++ member_type_name +++ "__encode" +++ " value))\n"

		// Utility functions

		clean_name = name

////////////////////////////////
//         ARRAYS
////////////////////////////////

compile_array :: CompilerState String TypeCode -> String
compile_array opt name (Tk_Array length tc)
	= foldl (+++) "" (map (\s -> (string_subst s clean_name)) components)
	where
		components = (if opt.genDef def_components impl_components)

		def_components
			= [gen_member_type, gen_comment,
			   gen_def, gen_tc_def, gen_decode_def, 
			   gen_encode_def]
		impl_components 
			= [gen_member_type, gen_comment, gen_def, 
			   gen_tc_def, gen_tc_impl, 
			   gen_decode_def, gen_decode_impl, gen_encode_def, gen_encode_impl]

		// Code generator functions
		// The '%' symbol is substituted for the name of the clean type

		gen_comment = gen_header_comment clean_name

		gen_member_type = if (is_anonym_member tc) (compile_anonym_type opt member_type_name tc) ""

		gen_unique = ""

		gen_def = "\n:: " +++ gen_unique +++ "% :== [" +++ (member_type_name) +++ "]\n"
		
		member_type_name = if (is_anonym_member tc) (clean_name +++ "__member_type") (clean_type_for tc)
		
		gen_tc_def = "%__tc :: TypeCode\n"

		gen_tc_impl = 
			"%__tc = Tk_Array " +++ (toString length) +++ " (" +++ (typecode_for tc member_type_name) +++ ")\n"

		gen_decode_def	= gen_decode_def_for clean_name

		gen_decode_impl	= 
			"%__decode (CorbaArray tc members) env\n" +++
			"\t| tc <> " +++ (typecode_for tc member_type_name) +++ "\n" +++
			"\t= (Nothing, env)\n" +++
			"\t| otherwise\n" +++
			"\t= map_maybe_with_env " +++ (member_type_name) +++ "__decode members env\n"																+++
			"%__decode _ env = (Nothing, env)\n"

		gen_encode_def	="%__encode :: % -> Any\n"

		gen_encode_impl	= "%__encode value = (CorbaArray %__tc (map " +++ member_type_name +++ "__encode" +++ " value))\n"

		// Utility functions

		clean_name = name

////////////////////////////////
//       REPOSITORIES
////////////////////////////////

compile_repository opt members
	= concat_list (map (\o -> compile_ir opt o) members)

////////////////////////////////
//        ENUMS
////////////////////////////////

compile_enum opt desc
	= foldl (+++) "" (map (\s -> (string_subst s clean_name)) components)
	where
		components = (if opt.genDef def_components impl_components)

		def_components 
			= [gen_comment, gen_def, gen_tc_def, gen_repoid, gen_name,
			   gen_decode_def, gen_encode_def, gen_equal_def, 
			   gen_toString_def]

		impl_components 
			= [gen_comment,
			   gen_def, gen_tc_def, gen_tc_impl, gen_repoid, gen_name, 
			   gen_names, gen_decode_def, gen_decode_impl, 
			   gen_encode_def, gen_encode_impl,
			   gen_equal_impl, gen_toString_impl]

		// Code Generation Functions

		gen_comment = gen_header_comment clean_name

		gen_def = "\n:: % = " +++ (toStringWith values "\n\t| ") +++ "\n"
		
		gen_tc_def	= "%__tc :: TypeCode\n"

		gen_tc_impl = 
			"%__tc = Tk_Enum %__repoid %__name %__memberNames\n"

		gen_repoid 	= "%__repoid \t:==\t \"" +++ repoid +++ "\"\n"
		gen_name	= "%__name \t:==\t \"" +++ name +++ "\"\n"

		gen_names	= "%__memberNames = [\n\t" +++ (toStringWith (map stringify names) "\n\t,") +++ "\n\t]" +++ "\n"

		gen_decode_def	= gen_decode_def_for clean_name

		gen_decode_impl	= 
			"%__decode (CorbaEnum tc value) env\n" +++
			"\t| tc <> %__tc\n" +++
			"\t= (Nothing, env)\n" +++
			"\t| otherwise\n" +++
			"\t= (decode_value value, env)\n" +++
			"\twhere\n" +++
			(toStringWith (map gen_member names) "\n") +++ "\n" +++
			"\t\tdecode_value _ \t=\t Nothing\n\n"				+++
			"%__decode _ env = (Nothing, env)\n"
		where
			gen_member name = "\t\tdecode_value " +++ (stringify name) +++ "\t=\t(Just " +++ (string_capitalize_first_char name) +++ ")"

		gen_encode_def	= "%__encode :: % -> Any\n"

		gen_encode_impl	= "%__encode value = (CorbaEnum %__tc (encode_value value))\nwhere\n" +++ (concat_list (map (\n -> ("\tencode_value " +++ (string_capitalize_first_char n) +++ "\t=\t" +++ (stringify n) +++ "\n")) names)) +++ "\n"

		gen_equal_def = "\ninstance == %\n"

		gen_equal_impl = "\ninstance == % where\n\t" +++ (toStringWith (map gen_value values) "\t") +++ "\t(==) _ _ = False\n"
		where
			gen_value value = "(==) " +++ value +++ " " +++ value +++ " = True\n"

		gen_toString_def = "\ninstance toString %\n"

		gen_toString_impl = "\ninstance toString % where\n\t" +++ (toStringWith (map gen_value names) "\t") +++ "\n"
		where
			gen_value name = "toString " +++ (string_capitalize_first_char name) +++ " = " +++ (stringify name) +++ "\n"

		// Utility Functions

		name   = desc.CORBA_TypeDescription.name

		repoid = desc.CORBA_TypeDescription.id

		clean_name = (gen_clean_name repoid)

		names = get_names desc.CORBA_TypeDescription.type
		where
			get_names (Tk_Enum repoid name members) = members

		values = map string_capitalize_first_char names
		
////////////////////////////////
//         STRUCTS
////////////////////////////////

compile_struct opt desc included_defs
	= foldl (+++) "" (map (\s -> (string_subst s clean_name)) components) +++ gen_included_defs
	where
		components = (if opt.genDef def_components impl_components)

		def_components 
			= [gen_comment,
			   gen_anonym_members, gen_def, gen_tc_def, gen_repoid, 
			   gen_name, gen_decode_def, gen_encode_def]

		impl_components 
			= [gen_anonym_members, gen_def, 
			   gen_tc_def, gen_tc_impl, gen_repoid, gen_name, 
			   gen_decode_def, gen_decode_impl, 
			   gen_encode_def, gen_encode_impl]

		// Code generator functions
		// The '%' symbol is substituted for the name of the clean type

		gen_included_defs = concat_list (map (\o -> compile_ir opt o) included_defs)

		gen_comment = gen_header_comment clean_name

		gen_anonym_members = toStringWith (map gen_member (filter (is_anonym_member o snd) members)) "\n"
		where
			gen_member (n, tc) = compile_anonym_type opt (clean_type_for_anonym_member clean_name n) tc

		gen_unique = ""

		gen_def = "\n:: " +++ (gen_unique) +++ "% = {\n\t" +++ (gen_members members) +++ "\n\t}\n"
		where
			gen_members []	= "foo :: Int"
			gen_members l 	= toStringWith (map gen_member l) ",\n\t"
			gen_member (n, tc) = n +++ " :: " +++ (type_for_member n tc)

		gen_tc_def	= "%__tc :: TypeCode\n"

		gen_tc_impl = 
			"%__tc = Tk_Struct %__repoid %__name " +++ "[\n\t" +++ (toStringWith (map gen_member members) ",\n\t") +++ "\n\t]" +++ "\n"
		where
			gen_member (name, tc) = "(" +++ (stringify name) +++ ", " +++ (typecode_for_member name tc) +++ ")"

		gen_repoid 	= "%__repoid \t:==\t \"" +++ repoid +++ "\"\n"
		gen_name	= "%__name \t:==\t \"" +++ type_name +++ "\"\n"

		gen_decode_def	= gen_decode_def_for clean_name

		gen_decode_impl =
			"%__decode (CorbaStruct tc members) env\n" +++
			"\t| tc <> %__tc\n" 						+++
			"\t= (Nothing, env)\n" 						+++
			"\t| otherwise\n" 							+++
			"\t= decode_struct\n"						+++
			"\twhere\n"									+++
			"\t\tdecode_struct\n"						+++
			"\t\t" +++ (toStringWith (map gen_member1 (zip2 members (indexList members))) "\n\t\t") +++ "\n"							+++
			"\t\t| and [" +++ (toStringWith (map (\i -> "is_just m" +++ (toString i)) (indexList members)) ", ") +++ "]\n" 						+++
			"\t\t= (Just { % |\n\t\t\t" +++ (gen_members members) +++ "\n\t\t\t}, env)\n" 													+++
			"\t\t| otherwise\n"							+++
			"\t\t= (Nothing, env)\n"					+++
			"%__decode _ env = (Nothing, env)\n"
		where
			gen_member1 ((name, tc), index) = "# (m" +++ (toString index) +++ ", env) = " +++ (decode_for_member name tc) +++ " (snd (members!!" +++ (toString index) +++ ")) env"
			gen_members [] = "foo = 0"
			gen_members l = toStringWith (map gen_member (zip2 members (indexList members))) ",\n\t\t\t"
			gen_member ((name, tc), index) = name +++ "\t=\t just m" +++ (toString index)

		gen_encode_def	= "%__encode :: % -> Any\n"

		gen_encode_impl	= "%__encode rec = (CorbaStruct %__tc [\n\t" +++ (toStringWith (map gen_member members) ",\n\t") +++ "\n\t])\n"
		where
			gen_member (name, tc) = "(" +++ (stringify name) +++ ", " +++ (encode_for_member name tc) +++ " rec.%." +++ name +++ ")"

		// Utilities

		type_name = desc.CORBA_TypeDescription.name

		repoid = desc.CORBA_TypeDescription.id

		clean_name = (gen_clean_name repoid)

		members = get_names desc.CORBA_TypeDescription.type
		where
			get_names (Tk_Struct repoid name members) = members

		// Handle anonymous members
		type_for_member :: String TypeCode -> String
		type_for_member name tc = if (is_anonym_member tc) (clean_type_for_anonym_member clean_name name) (clean_type_for tc)

		// Handle anonymous members
		typecode_for_member :: String TypeCode -> String
		typecode_for_member name tc = typecode_for tc (clean_type_for_anonym_member clean_name name)

		// Handle anonymous members
		decode_for_member :: String TypeCode -> String
		decode_for_member name tc = if (is_anonym_member tc) ((clean_type_for_anonym_member clean_name name) +++ "__decode") (decode_for tc)

		// Handle anonymous members
		encode_for_member :: String TypeCode -> String
		encode_for_member name tc = if (is_anonym_member tc) ((clean_type_for_anonym_member clean_name name) +++ "__encode") (encode_for tc)

////////////////////////////////
//         UNIONS
////////////////////////////////

compile_union opt desc included_defs
	= foldl (+++) "" (map (\s -> (string_subst s clean_name)) components) +++ gen_included_defs
	where
		components = (if opt.genDef def_components impl_components)

		def_components 
			= [gen_comment, gen_anonym_members, gen_def, gen_tc_def, 
			   gen_repoid, gen_name, gen_decode_def, gen_encode_def]

		impl_components 
			= [gen_anonym_members, gen_def, 
			   gen_tc_def, gen_tc_impl, gen_repoid, gen_name, 
			   gen_decode_def, gen_decode_impl, 
			   gen_encode_def, gen_encode_impl]

		// Code generator functions
		// The '%' symbol is substituted for the name of the clean type

		gen_included_defs = concat_list (map (\o -> compile_ir opt o) included_defs)

		gen_comment = gen_header_comment clean_name

		gen_anonym_members = toStringWith (map gen_member (filter (\(n,d,t) -> is_anonym_member t) members)) "\n"
		where
			gen_member (n, disc, tc) = compile_anonym_type opt (clean_type_for_anonym_member clean_name n) tc

		gen_unique = ""

		gen_def = "\n\n:: " +++ gen_unique +++ "% = \n\t" +++ (toStringWith (map gen_member cooked_members) "\n\t| ") +++ "\n\n"
		where
			gen_member member=:(n, disc, tc)
				| is_implicit_member member
				= (name_for_discriminator disc) +++ " " +++ (clean_type_for disc_type)
				| is_default_member (n, disc, tc)
				= (name_for_discriminator disc) +++ " " +++ (clean_type_for disc_type) +++ " " +++ (type_for_member n tc)
				| otherwise
				= (name_for_discriminator disc) +++ " " +++ (type_for_member n tc)

		gen_tc_def	= "%__tc :: TypeCode\n"

		gen_tc_impl = 
			"%__tc = Tk_Union %__repoid %__name " +++ (typecode_for disc_type "") +++ " [\n\t" +++ (toStringWith (map gen_member members) ",\n\t") +++ "\n\t]" +++ "\n"
		where
			gen_member (name, disc, tc) = "(" +++ (stringify name) +++ ", " +++ (clean_encoding_for disc) +++ ", " +++ (typecode_for_member name tc) +++ ")"

		gen_repoid 	= "%__repoid \t:==\t \"" +++ repoid +++ "\"\n"
		gen_name	= "%__name \t:==\t \"" +++ type_name +++ "\"\n"

		gen_decode_def	= gen_decode_def_for clean_name

		gen_decode_impl	= 
			"%__decode (CorbaUnion tc discriminator value) env\n"	+++
			"\t| tc <> %__tc\n"										+++
			"\t= (Nothing, env)\n"									+++
			"\t| otherwise\n"										+++
			"\t= decode discriminator env\n"						+++
			"\twhere\n"												+++
			"\t\t" +++ (toStringWith (map gen_member cooked_members) "\n\t\t") +++ "\n" 															+++
			"%__decode _ env = (Nothing, env)\n"
		where
			gen_member member=:(name, disc, tc)
				// The cooked_members function guarantees that
				// the default member is generated last
				| is_implicit_member member =
				"decode disc env\n"										+++
				"\t\t\t# (d, env) = " +++ (clean_type_for disc_type) +++ "__decode disc env\n"															+++
				"\t\t\t| is_just d = (Just (" +++ (name_for_discriminator disc) +++ " (just d)), env)\n" 							+++
				"\t\t\t| otherwise = (Nothing, env)"
				| is_default_member member =
				"decode disc env\n"										+++
				"\t\t\t# (d, env) = " +++ (clean_type_for disc_type) +++ "__decode disc env\n"															+++
				"\t\t\t# (v, env) = " +++ (decode_for_member name tc) +++ " value env\n"																+++
				"\t\t\t| is_just v && is_just d = (Just (" +++ (name_for_discriminator disc) +++ " (just d) (just v)), env)\n" 							+++
				"\t\t\t| otherwise = (Nothing, env)"
				| otherwise =
				"decode " +++ (clean_encoding_for disc) +++ " env\n"	+++
				"\t\t\t# (v, env) = " +++ (decode_for_member name tc) +++ " value env\n"																+++
				"\t\t\t| is_just v = (Just (" +++ (name_for_discriminator disc) +++ " (just v)), env)\n" 												+++
				"\t\t\t| otherwise = (Nothing, env)"

		gen_encode_def	= clean_name +++ "__encode :: " +++ clean_name +++ " -> Any\n"

		gen_encode_impl	= clean_name +++ "__encode rec = encode rec\nwhere\n\t" +++ (toStringWith (map gen_member cooked_members) "\t") +++ "\n"
		where
			gen_member member=:(name, disc, tc)
				| is_implicit_member member
				= "encode (" +++ (name_for_discriminator disc) +++ " disc) = (CorbaUnion " +++ clean_name +++ "__tc (" +++ (clean_type_for disc_type) +++ "__encode disc) (CorbaShort 0))\n"
				| is_default_member member
				= "encode (" +++ (name_for_discriminator disc) +++ " disc value) = (CorbaUnion " +++ clean_name +++ "__tc (" +++ (clean_type_for disc_type) +++ "__encode disc) (" +++ (encode_for_member name tc) +++ " value))\n"
				| otherwise
				= "encode (" +++ (name_for_discriminator disc) +++ " value) = (CorbaUnion " +++ clean_name +++ "__tc " +++ (clean_encoding_for disc) +++ " (" +++ (encode_for_member name tc) +++ " value))\n"

		// Utilities

		type_name = desc.CORBA_TypeDescription.name

		repoid = desc.CORBA_TypeDescription.id

		clean_name = (gen_clean_name repoid)

		(disc_type, members) = get_info desc.CORBA_TypeDescription.type
		where
			get_info (Tk_Union repoid name tc members) = (tc, members)

		//
		// Construct a new list of members as follows:
		//  - if there is a default member, it is placed at the end of the list
		//  - if there is no default member, an implicit default member is added
		//    to the end of the list
		//

		cooked_members
			| not has_default_member
			= members ++ [("foo", (CorbaOctet 0), Tk_Null)]
			| otherwise
			= 
			(filter (not o is_default_member) members) ++
			(filter is_default_member members)

		name_for_discriminator disc = type_name +++ "_" +++ disc_name
		where 
			disc_name = case disc of
				(CorbaShort i) 		-> toString i
				(CorbaLong i)  		-> toString i
				(CorbaUShort i) 	-> toString i
				(CorbaULong i)		-> toString i
				(CorbaChar i)		-> abort "union discriminator type Char is not supported."
				(CorbaOctet i)		-> "default"
				(CorbaBoolean b)	-> toString b
				(CorbaEnum tc s)	-> s

		is_default_member (_, (CorbaOctet i), _) 	= True
		is_default_member _				 			= False

		is_implicit_member :: (String, Any, TypeCode) -> Bool
		is_implicit_member (_, (CorbaOctet i), Tk_Null)	=	True
		is_implicit_member _								=	False

		has_default_member = (length (filter is_default_member members)) > 0

		// Handle anonymous members
		type_for_member :: String TypeCode -> String
		type_for_member name tc = if (is_anonym_member tc) (clean_type_for_anonym_member clean_name name) (clean_type_for tc)

		// Handle anonymous members
		typecode_for_member :: String TypeCode -> String
		typecode_for_member name tc = typecode_for tc (clean_type_for_anonym_member clean_name name)

		// Handle anonymous members
		decode_for_member :: String TypeCode -> String
		decode_for_member name tc = if (is_anonym_member tc) ((clean_type_for_anonym_member clean_name name) +++ "__decode") (decode_for tc)

		// Handle anonymous members
		encode_for_member :: String TypeCode -> String
		encode_for_member name tc = if (is_anonym_member tc) ((clean_type_for_anonym_member clean_name name) +++ "__encode") (encode_for tc)

////////////////////////////////
//        EXCEPTIONS
////////////////////////////////

compile_exception opt desc
	= foldl (+++) "" (map (\s -> (string_subst s clean_name)) components)
	where
		components = (if opt.genDef def_components impl_components)

		def_components 
			= [gen_comment,
			   gen_anonym_members, gen_def, gen_tc_def, gen_repoid, 
			   gen_name, gen_decode_def, gen_encode_def]

		impl_components 
			= [gen_anonym_members, gen_def, 
			   gen_tc_def, gen_tc_impl, gen_repoid, gen_name, 
			   gen_decode_def, gen_decode_impl, 
			   gen_encode_def, gen_encode_impl]

		// Code generator functions
		// The '%' symbol is substituted for the name of the clean type

		gen_comment = gen_header_comment clean_name

		gen_anonym_members = toStringWith (map gen_member (filter (is_anonym_member o snd) members)) "\n"
		where
			gen_member (n, tc) = compile_anonym_type opt (clean_type_for_anonym_member clean_name n) tc

		gen_unique = ""

		gen_def = "\n:: " +++ (gen_unique) +++ "% = {\n\t" +++ (gen_members members) +++ "\n\t}\n"
		where
			gen_members []	= "foo :: Int"
			gen_members l 	= toStringWith (map gen_member l) ",\n\t"
			gen_member (n, tc) = n +++ " :: " +++ (type_for_member n tc)

		gen_tc_def	= "%__tc :: TypeCode\n"

		gen_tc_impl = 
			"%__tc = Tk_Exception %__repoid %__name " +++ "[\n\t" +++ (toStringWith (map gen_member members) ",\n\t") +++ "\n\t]" +++ "\n"
		where
			gen_member (name, tc) = "(" +++ (stringify name) +++ ", " +++ (typecode_for_member name tc) +++ ")"

		gen_repoid 	= "%__repoid \t:==\t \"" +++ repoid +++ "\"\n"
		gen_name	= "%__name \t:==\t \"" +++ type_name +++ "\"\n"

		gen_decode_def	= gen_decode_def_for clean_name

		gen_decode_impl =
			"%__decode (CorbaException tc members) env\n" +++
			"\t| tc <> %__tc\n" 						+++
			"\t= (Nothing, env)\n" 						+++
			"\t| otherwise\n" 							+++
			"\t= decode_struct\n"						+++
			"\twhere\n"									+++
			"\t\tdecode_struct\n"						+++
			"\t\t" +++ (toStringWith (map gen_member1 (zip2 members (indexList members))) "\n\t\t") +++ "\n"							+++
			"\t\t| and [" +++ (toStringWith (map (\i -> "is_just m" +++ (toString i)) (indexList members)) ", ") +++ "]\n" 						+++
			"\t\t= (Just { % |\n\t\t\t" +++ (gen_members members) +++ "\n\t\t\t}, env)\n" 													+++
			"\t\t| otherwise\n"								+++
			"\t\t= (Nothing, env)\n"						+++
			"%__decode _ env = (Nothing, env)\n"
		where
			gen_member1 ((name, tc), index) = "# (m" +++ (toString index) +++ ", env) = " +++ (decode_for_member name tc) +++ " (snd (members!!" +++ (toString index) +++ ")) env"
			gen_members [] = "foo = 0"
			gen_members l = toStringWith (map gen_member (zip2 members (indexList members))) ",\n\t\t\t"
			gen_member ((name, tc), index) = name +++ "\t=\t just m" +++ (toString index)

		gen_encode_def	= "%__encode :: % -> Any\n"

		gen_encode_impl	= "%__encode rec = (CorbaException %__tc [\n\t" +++ (toStringWith (map gen_member members) ",\n\t") +++ "\n\t])\n"
		where
			gen_member (name, tc) = "(" +++ (stringify name) +++ ", " +++ (encode_for_member name tc) +++ " rec.%." +++ name +++ ")"

		// Utilities

		type_name = desc.CORBA_ExceptionDescription.name

		repoid = desc.CORBA_ExceptionDescription.id

		clean_name = (gen_clean_name repoid)

		members = get_names desc.CORBA_ExceptionDescription.type
		where
			get_names (Tk_Exception repoid name members) = members

		// Handle anonymous members
		type_for_member :: String TypeCode -> String
		type_for_member name tc = if (is_anonym_member tc) (clean_type_for_anonym_member clean_name name) (clean_type_for tc)

		// Handle anonymous members
		typecode_for_member :: String TypeCode -> String
		typecode_for_member name tc = typecode_for tc (clean_type_for_anonym_member clean_name name)

		// Handle anonymous members
		decode_for_member :: String TypeCode -> String
		decode_for_member name tc = if (is_anonym_member tc) ((clean_type_for_anonym_member clean_name name) +++ "__decode") (decode_for tc)

		// Handle anonymous members
		encode_for_member :: String TypeCode -> String
		encode_for_member name tc = if (is_anonym_member tc) ((clean_type_for_anonym_member clean_name name) +++ "__encode") (encode_for tc)


////////////////////////////////
//        TYPE ALIASES
////////////////////////////////

compile_alias opt desc
	= case desc.CORBA_TypeDescription.type of

	// sequences can only be defined using an alias
	(Tk_Alias repoid name tc=:(Tk_Sequence _ _)) -> compile_sequence opt (gen_clean_name repoid) tc

	// arrays can only be defined using an alias
	(Tk_Alias repoid name tc=:(Tk_Array _ _)) -> compile_array opt (gen_clean_name repoid) tc

	// a real alias for an already generated type
	(Tk_Alias repoid name tc) -> foldl (+++) "" components
	where
		components = (if opt.genDef def_components impl_components)

		def_components = [gen_comment, gen_def, gen_tc_def, gen_repoid, 
						  gen_name, gen_decode, gen_encode]

		impl_components = [gen_comment,
						   gen_def, gen_tc_def, gen_tc_impl, gen_repoid, 
						   gen_name, gen_decode, gen_encode]

		// Code generator functions

		gen_comment = gen_header_comment clean_name

		gen_unique = ""

		gen_def = "\n:: " +++ (gen_unique) +++ clean_name +++ " :== " +++ aliased_name +++ "\n"

		gen_tc_def	= clean_name +++ "__tc :: TypeCode\n"

		gen_tc_impl = clean_name +++ "__tc \t=\t " +++ aliased_name +++ "__tc\n"

		gen_repoid 	= clean_name +++ "__repoid \t:==\t \"" +++ repoid +++ "\"\n"
		gen_name	= clean_name +++ "__name \t:==\t \"" +++ name +++ "\"\n"

		gen_decode	= clean_name +++ "__decode :== " +++ aliased_name +++ "__decode\n"

		gen_encode	= clean_name +++ "__encode :== " +++ aliased_name +++ "__encode\n"

		// Utility functions

		clean_name = (gen_clean_name repoid)

		aliased_name = (clean_type_for tc)

////////////////////////////////
//        MODULES
////////////////////////////////

compile_module opt desc members 
	= concat_list (map (\o -> compile_ir opt o) members)

////////////////////////////////
//       CONSTANTS
////////////////////////////////

compile_constant opt desc
	= "\n" +++ clean_name +++ " :== " +++ clean_value +++ "\n"
where
	name = desc.CORBA_ConstantDescription.id
	type = desc.CORBA_ConstantDescription.type
	value = desc.CORBA_ConstantDescription.value
	
	clean_name = gen_clean_name name
	clean_value = clean_value_for value

////////////////////////////////
//      INTERFACES
////////////////////////////////

compile_interface opt desc members
	= foldl (+++) "" (map (\s -> (string_subst s clean_name)) components) +++ gen_members
	where
		components = (if opt.genDef def_components impl_components)

		def_components 
			= [gen_comment, gen_def, gen_tc_def, gen_repoid, gen_name,
			   gen_decode_def, gen_encode_def, gen_nil_def,
			   gen_narrow_def, gen_widen, gen_widen_instances_def,
			   gen_servant_def, gen_servant_open_def]

		impl_components 
			= [gen_def, gen_tc_def, gen_tc_impl, 
			   gen_repoid, gen_name, gen_create_def, gen_create_impl,
			   gen_decode_def, gen_decode_impl, gen_encode_def, 
			   gen_encode_impl, gen_nil_def, gen_nil_impl,
			   gen_narrow_impl, gen_widen, gen_widen_instances_impl,
			   gen_servant_def, gen_servant_open_def, gen_servant_open_impl]

		// Code generator functions
		// The '%' symbol is substituted for the name of the clean type

		gen_comment = gen_header_comment "%"

		gen_def = "\n:: %  = { reference :: CORBA_Object }\n"

		gen_tc_def 	= "%__tc \t :: TypeCode\n"

		gen_tc_impl = "%__tc \t=\t (Tk_ObjRef %__repoid %__name)\n"

		gen_repoid 	= "%__repoid \t:==\t " +++ (stringify repoid) +++ "\n"
		gen_name	= "%__name \t:==\t " +++ (stringify name) +++ "\n"

		gen_create_def = "%__create :: CORBA_Object -> %\n"

		gen_create_impl = "%__create ref = { % | reference = ref }\n"

		gen_decode_def	= gen_decode_def_for "%"

		gen_decode_impl	= 
			"%__decode (CorbaObjRef ref) env = (Just (%__create ref), env)\n" +++
			"%__decode _ env = (Nothing, env)\n"

		gen_encode_def  = "%__encode :: % -> Any\n"

		gen_encode_impl  = "%__encode obj = (CorbaObjRef obj.%.reference)\n"

		gen_nil_def = "%__nil :: *World -> (%, *World)\n"

		gen_nil_impl = "%__nil env\n\t# (ref, env) = CORBA_Object__nil env\n\t= ({ % | reference = ref }, env)\n"

		gen_narrow_def = "class %__narrow_class a where\n" +++ "\t%__narrow :: a *World -> (MayBe %, *World)\n" +++ (toStringWith (map gen_instance base_interfaces) "")
		where
			gen_instance base_repoid = "instance %__narrow_class " +++ base_name +++ "\n"
			where
				base_name = (gen_clean_name base_repoid)		

		gen_narrow_impl = "class %__narrow_class a where\n" +++ "\t%__narrow :: a *World -> (MayBe %, *World)\n" +++ (toStringWith (map gen_instance base_interfaces) "")
		where
			gen_instance base_repoid = "instance %__narrow_class " +++ base_name +++ " where\n\t%__narrow obj env\n\t# (is_a, env) = CORBA_Object_is_a (CORBA_Object__widen obj) %__repoid env\n\t| is_a\n\t= " +++ "(Just { % | reference = (CORBA_Object__widen obj) }, env)\n\t| otherwise\n\t= (Nothing, env)\n"
			where
				base_name = (gen_clean_name base_repoid)		

		// The overloaded widen operation is defined here, while
		// instances for derived interfaces are defined with that interface
		gen_widen = "\nclass %__widen_class a where\n\t%__widen :: a -> %\n"

		// Generate instances of the widening functions in base interfaces
		gen_widen_instances_def = toStringWith (map gen_instance base_interfaces) ""
		where
			gen_instance base_repoid 
				= "\ninstance " +++ base_name +++ "__widen_class %\n"
			where
				base_name = (gen_clean_name base_repoid)

		// Generate instances of the widening functions in base interfaces
		gen_widen_instances_impl = toStringWith (map gen_instance base_interfaces) ""
		where
			gen_instance base_repoid = "\ninstance " +++ base_name +++ "__widen_class % where\n\t" +++ base_name +++ "__widen { % | reference } = " +++ base_name +++ "__create reference\n"
			where
				base_name = (gen_clean_name base_repoid)

		gen_members = concat_list (map (\o -> compile_ir opt o) members)

		gen_servant_def = "\n:: %__servant l = E.a: {\n\t ls :: a" +++
				(toStringWith2
					(map gen_op_def operations) ",\n\t ") +++ "\n\t };\n"
		where
			gen_op_def op 	= compile_operation {opt & genServantDef = True} (new_op op)
			new_op op		= {CORBA_OperationDescription | op & name = "impl_" +++ op.CORBA_OperationDescription.name}

		gen_servant_open_def = "\n%__servant_open :: (PSt l) (%__servant l) *World -> (CORBA_Object, (PSt l), *World)\n"

		gen_servant_open_impl = 
			"\n%__servant_open ps servant env\n\t = CORBA_Server_add_servant ps " +++ (stringify repoid) +++ " serv env\nwhere\n\tserv = { CORBA_Servant_Wrapper | obj = servant, dispatch = dispatch_fun};\n" +++ gen_dispatch_fun +++ "\n" +++ gen_wrappers +++ "\n"
		where
			gen_dispatch_fun = gen_dispatch_def +++ gen_dispatch_impl
			gen_dispatch_def = "\tdispatch_fun :: (PSt l) CORBA_ServerRequest (%__servant l) *World -> (((PSt l), (%__servant l)), *World)\n"
			gen_dispatch_impl = "\tdispatch_fun ps req s w\n\t\t# (op, w) = CORBA_ServerRequest_operation req w\n\t\t= case op of\n" +++ (toStringWith2 (map gen_line operations) "\t\t") +++ "\n"
			gen_line op = (stringify op_name) +++ "\t-> %__wrap_" +++ op_name +++ " ps req s w\n"
			where
				op_name = op.CORBA_OperationDescription.name
			gen_wrappers = toStringWith (map gen_wrapper operations) "\n"
			where
				gen_wrapper op = compile_operation {opt & genWrapper = True } op

		// Utility functions

		repoid 	=	desc.CORBA_InterfaceDescription.id

		name	=	desc.CORBA_InterfaceDescription.name

		base_interfaces = desc.CORBA_InterfaceDescription.base_interfaces ++ ["IDL:CORBA/Object:1.0"]

		clean_name = (gen_clean_name repoid)

		// All the methods in the interface, plus the get/set methods
		// of the attributes
		operations = map (\(IROperation desc) -> desc) ops
		where
			ops 		= (filter is_operation members) ++ arg_ops
			arg_ops		= map (\d -> (IROperation d)) (flatten (map (\a -> get_operations_for a opt.idlType_nil) attributes))
			attributes 	= map (\(IRAttribute desc) -> desc) (filter is_attribute members)
			is_operation (IROperation _)	= True
			is_operation _					= False
			is_attribute (IRAttribute _)	= True
			is_attribute _					= False

////////////////////////////////
//       OPERATIONS
////////////////////////////////

compile_operation opt desc
	= foldl (+++) "" components
	where
		components
			| opt.genServantDef
			= servant_def_components
			| opt.genWrapper
			= wrapper_components
			| opt.genDef
			= def_components
			| otherwise
			= impl_components

		def_components			= [gen_def]
		impl_components 		= [gen_def, gen_impl]
		servant_def_components 	= [gen_servant_def]
		wrapper_components		= [gen_wrapper_def, gen_wrapper_impl]

		// Code Generation routines

		gen_def = "\n" +++ interface_name +++ "_" +++ op_name +++ " :: " +++ interface_name +++ " " +++ (toStringWith (map gen_arg input_args) " ") +++ " *World\n\t-> ((ResultOrException " +++ gen_return_type +++ " CorbaException), *World)\n"
		where
			gen_arg arg = (clean_type_for arg.CORBA_ParameterDescription.type)
			gen_return_type 
				= case has_output_args of
				False	->	 (clean_type_for return_type)
				True	->	 "(" +++ (clean_type_for return_type) +++ ", " +++ (toStringWith (map (\arg -> (clean_type_for arg.CORBA_ParameterDescription.type)) output_args) ", ") +++ ")"

		gen_impl = 
			interface_name +++ "_" +++ op_name +++ " obj " +++ (toStringWith (map (\arg -> "arg_" +++ arg.CORBA_ParameterDescription.name) input_args) " ") +++ " env\n" 															+++ 
			"\t# (result, exception, " +++ gen_args_variable +++ ", env) = corba_invoke obj." +++ interface_name +++ ".reference " +++ (stringify op_name) +++ gen_args +++ (typecode_for return_type "") +++ " [] env\n" +++ 
			gen_decode_return_value 						+++
			"\t= (corba_result_create " +++ gen_return_value +++ " exception, env)\n"
		where
			gen_args
			| length args == 0
			= "[]"
			| otherwise
			=  " [\n\t\t" +++ (toStringWith (map gen_arg args) ",\n\t\t") +++ "\n\t\t] " 

			gen_arg arg
			| mode == PARAM_OUT
			= "(" +++ (constructor mode) +++ " " +++ (stringify name) +++ " " +++ (typecode_for type "") +++ ")"
			| otherwise
			= "(" +++ (constructor mode) +++ " " +++ (stringify name) +++ " (" +++ (encode_for type) +++ " " +++ "arg_" +++ name +++ "))"
			where
				name = arg.CORBA_ParameterDescription.name 
				type = arg.CORBA_ParameterDescription.type
				mode = arg.CORBA_ParameterDescription.mode
				constructor m = case m of
					PARAM_IN	->	"InArg"
					PARAM_OUT	->	"OutArgTemplate"
					PARAM_INOUT	->	"InOutArg"

			gen_args_variable = if has_output_args "args" "_"

			gen_decode_return_value 
				= case has_output_args of
				False	->	"\t# (result, env) = " +++ (decode_for return_type) +++ " result env\n"
				True	->	"\t# (result, env) = " +++ (decode_for return_type) +++ " result env\n" +++ "\t" +++ (toStringWith (map gen_arg (zip2 output_args [0..])) "\n\t") +++ "\n"
			where
				gen_arg (arg, index) = "# (r" +++ (toString index) +++ ", env) = " +++ (decode_for arg.CORBA_ParameterDescription.type) +++ " (corba_arg_get_value (args!!" +++ (toString index) +++ ")) env"

			gen_return_value 
				= case has_output_args of
				False	->	"(just result)"
				True	->	"((just result), " +++ (toStringWith (map gen_arg (zip2 output_args [0..])) ", ") +++ ")"
			where
				gen_arg (arg, index) = "(just r" +++ (toString index) +++ ")"

		gen_servant_def = op_name +++ " :: (a, (PSt l)) " +++ (toStringWith (map gen_arg input_args) " ") +++ " *World\n\t\t-> ((a, (PSt l)), (ResultOrException " +++ gen_return_type +++ " CorbaException), *World)"
		where
			gen_arg arg = (clean_type_for arg.CORBA_ParameterDescription.type)
			gen_return_type 
				= case has_output_args of
				False	->	 (clean_type_for return_type)
				True	->	 "(" +++ (clean_type_for return_type) +++ ", " +++ (toStringWith (map (\arg -> (clean_type_for arg.CORBA_ParameterDescription.type)) output_args) ", ") +++ ")"

		gen_wrapper_def = "\n" +++ interface_name +++ "__wrap_" +++ op_name +++ " :: (PSt l) CORBA_ServerRequest (" +++ interface_name +++ "__servant l) *World -> (((PSt l), (" +++ interface_name +++ "__servant l)), *World)\n"

		gen_wrapper_impl = 
			"\n" +++ interface_name +++ "__wrap_" +++ op_name +++ " ps req servant=:{ls, impl_" +++ op_name +++ "} w\n"						+++
			"\t# w\t= CORBA_ServerRequest_set_arg_types req " +++ gen_arg_types +++ " w\n"												+++
			gen_decode_args									+++
			gen_call										+++
			gen_send_result
		where
			gen_arg_types
			| length args == 0
			= "[]"
			| otherwise
			=  " [\n\t\t" +++ (toStringWith (map gen_arg args) ",\n\t\t") +++ "\n\t\t] " 
			where
				gen_arg arg
				= "(" +++ (constructor mode) +++ " " +++ (stringify name) +++ " " +++ (typecode_for type "") +++ ")"
				where
					name = arg.CORBA_ParameterDescription.name 
					type = arg.CORBA_ParameterDescription.type
					mode = arg.CORBA_ParameterDescription.mode
					constructor m = case m of
						PARAM_IN	->	"InArgTemplate"
						PARAM_OUT	->	"OutArgTemplate"
						PARAM_INOUT	->	"InOutArgTemplate"

			gen_decode_args
			| length args == 0
			= ""
			| otherwise
			= (toStringWith (map gen_arg indexed_input_args) "\n") +++ "\n"
			where
				gen_arg (arg, index)
				= string_subst ("\t# (arg%, w)\t= CORBA_ServerRequest_get_arg req % w\n\t# (arg%, w) = " +++ (decode_for type) +++ " arg% w\n\t# arg%    = just arg%") (toString index)
				where
					name = arg.CORBA_ParameterDescription.name 
					type = arg.CORBA_ParameterDescription.type
					mode = arg.CORBA_ParameterDescription.mode

			gen_call
				= "\t# ((ls, ps), result, w) = impl_" +++ op_name +++ " (ls, ps) " +++ gen_args +++ " w\n"
			where
				gen_args 
					= toStringWith (map arg_name_for indexed_input_args) " "
			
			gen_send_result
				= "\t# w\t= " +++ aux_fun_name +++ " req result w\n\t= ((ps, {" +++ interface_name +++ "__servant | servant & ls=ls}), w)\n\twhere\n\t\t" +++ gen_result_case +++ gen_exception_case
			where
				gen_result_case
					= aux_fun_name +++ " req (Result " +++ gen_arglist  +++ ") w" +++ gen_encode_args +++ "\n" +++ gen_store_result +++ "\n"

				gen_exception_case
					= "\t\t" +++ aux_fun_name +++ " req (Exception ex) w \n\t\t= CORBA_ServerRequest_set_exception req (corba_exception_encode ex) w\n"

				aux_fun_name = interface_name +++ "__wrapaux_" +++ op_name

				gen_arglist
					| has_output_args
					= "(res, " +++ (toStringWith (map arg_name_for indexed_output_args) ",") +++ ")"
					| otherwise
					= "res"
				gen_encode_args
					= toStringWith2 (map gen_arg indexed_output_args) "\n"
				where
					gen_arg (arg, index)
						= "\t\t# w = CORBA_ServerRequest_set_arg req " +++ (toString index) +++ " (" +++ (encode_for type) +++ " " +++ (arg_name_for (arg, index)) +++ ") w"
					where
						name = arg.CORBA_ParameterDescription.name 
						type = arg.CORBA_ParameterDescription.type
						mode = arg.CORBA_ParameterDescription.mode

				gen_store_result
					| has_return_value
					= "\t\t= CORBA_ServerRequest_set_result req (" +++ (encode_for return_type) +++ " res) w"
					| otherwise
					= "\t\t= w"

		// Utility routines

		interface_name	= (gen_clean_name desc.CORBA_OperationDescription.defined_in)
		op_name 		= desc.CORBA_OperationDescription.name

		return_type 	= desc.CORBA_OperationDescription.result

		args 			= desc.CORBA_OperationDescription.parameters

		input_args		= filter is_input_arg args

		output_args		= filter is_output_arg args

		is_input_arg arg	= arg.CORBA_ParameterDescription.mode <> PARAM_OUT
		is_output_arg arg	= arg.CORBA_ParameterDescription.mode <> PARAM_IN

		has_output_args = (length output_args) > 0

		indexed_input_args = filter (\(a, i) -> is_input_arg a) (with_indexes args)

		indexed_output_args = filter (\(a, i) -> is_output_arg a) (with_indexes args)

		arg_name_for (arg, index) = "arg" +++ (toString index)

		has_return_value = has_return_value2 return_type
		where
			has_return_value2 (Tk_Void) 	= False
			has_return_value2 _				= True

////////////////////////////////
//       ATTRIBUTES
////////////////////////////////

compile_attribute opt desc
	// Convert it into an operation invocation
	= toStringWith (map (compile_operation opt) ops) ""
	where
		ops = get_operations_for desc opt.idlType_nil	

////////////////////////////////
//         FRONT END
////////////////////////////////

compile_ir_def opt obj = compile_ir opt obj

gen_def_file :: CompilerState String IRObject -> String
gen_def_file opt modulename obj = "\ndefinition module " +++ modulename +++ "\n\nimport StdEnv, Corba\n\n" +++ (compile_ir_def opt obj) +++ "\n"

compile_ir_impl opt obj = compile_ir opt obj

gen_impl_file :: CompilerState String IRObject -> String
gen_impl_file opt modulename obj = "\nimplementation module " +++ modulename +++ "\n\nimport StdEnv, Corba\n\n" +++ (compile_ir_impl opt obj) +++ "\n"

compile_idl :: String IRObject CORBA_IDLType *Files -> *Files
compile_idl modulename ir idlType_nil files
	# (open_ok, def_file, files)	=	fopen (modulename +++ ".dcl") FWriteText files
	| not open_ok
	= abort "Unable to create output file."
	# (open_ok, impl_file, files)	=	fopen (modulename +++ ".icl") FWriteText files
	| not open_ok
	= abort "Unable to create output file."
	# def_file	=	fwrites (gen_def_file def_options modulename ir) def_file
	# impl_file	=	fwrites (gen_impl_file impl_options modulename ir) impl_file
	# (_, files) = fclose def_file files
	# (_, files) = fclose impl_file files
	= files
where
	def_options = { CompilerState | 	
						genDef 			= True, 
						idlType_nil 	= idlType_nil,
						genServantDef 	= False,
						genWrapper		= False
				 	}
	impl_options = { CompilerState | 
						genDef 			= False,
						idlType_nil		= idlType_nil,
						genServantDef	= False,
						genWrapper		= False
					}

////////////////////////////////
//          MAIN
////////////////////////////////

Start w
	| argc <> 2
	= abort "This program must be run from the idl2clean script.\n"
	# (orb, _, w)			= CORBA_ORB_init ["foo"] w
	# (obj, w) 				= resolve_ir orb w
	# (Just repository, w)	= CORBA_IRObject__narrow obj w
	# (ir, w)	 			= ir_dump repository w
	# (idlType_nil, w)		= CORBA_IDLType__nil w
	= appFiles (compile_idl module_name ir idlType_nil) w
	where
		args 	= getCommandLine
		argc 	= size args - 1

		repository_addr	 	=	args.[1]
//		repository_addr		=	"corbaloc::localhost:4141/DefaultRepository"
		module_name			=	args.[2]

		resolve_ir orb env
		# (obj, env) = CORBA_ORB_bind orb "IDL:omg.org/CORBA/Repository:1.0" repository_addr env
//		# (obj, env) = CORBA_ORB_string_to_object orb repository_addr env
//		# (obj, env) = CORBA_ORB_resolve_initial_references orb "InterfaceRepository" env
		= (decode_or_die obj, env)
		where
			decode_or_die (Just obj) = obj
			decode_or_die (Nothing) = abort ("unable to bind to IR at address " +++ (repository_addr) +++ "\n")





// Local Variables:
// tab-width: 4
// End:

